/*
 *  Copyright (C) 2012 Yvon TANGUY
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.db;

import java.io.File;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Date;

import org.apache.http.client.CookieStore;
import org.apache.http.cookie.Cookie;
import org.apache.http.impl.client.BasicCookieStore;
import org.apache.http.impl.cookie.BasicClientCookie2;
import org.vono.narau.R;
import org.vono.narau.common.Common;
import org.vono.narau.models.history.OfflineResult;
import org.vono.narau.models.history.OnlineResult;
import org.vono.narau.models.history.SearchResult;
import org.vono.narau.preferences.Preferences;
import org.vono.narau.utils.Language;

import android.content.ContentValues;
import android.database.Cursor;
import android.database.sqlite.SQLiteDatabase;
import android.database.sqlite.SQLiteException;
import android.util.Log;

/**
 * Dictionary search history, store in the cache dir of the device.
 * 
 */
public class HistoryDB extends CommonDB {
	private static final String TAG = HistoryDB.class.getSimpleName();

	/* filename */
	static final String DB_FILENAME = "dictionary_search_history.db";

	/* tables */
	private static final String TABLE_OFFLINE_SEARCH = "offline_search";
	private static final String TABLE_OFFLINE_RESULTS = "offline_results";
	private static final String TABLE_ONLINE_SEARCH = "online_search";
	private static final String TABLE_ONLINE_COOKIES = "online_cookies";

	/* columns */
	/**
	 * "id_pk", "date", "term", "lang", "num_results"
	 */
	private static final String[] COLS_OFFLINE_SEARCH = { "id_pk", "date", "term", "lang", "num_results" };
	/**
	 * "id_fk", "id_fk_dict"
	 */
	private static final String[] COLS_OFFLINE_RESULT = { "id_fk", "id_fk_dict" };
	/**
	 * "id_pk", "date", "term", "num_result", "srclang", "dstlang", "service_name", "html_text"
	 */
	private static final String[] COLS_ONLINE_SEARCH = { "id_pk", "date", "term", "num_result", "srclang", "dstlang",
			"service_name", "html_text" };
	/**
	 * service_name", "domain", "path", "name", "value", "expiry_date", "version
	 * ", "secure", "comment", "comment_url"
	 */
	private static final String[] COLS_ONLINE_COOKIES = { "service_name", "domain", "path", "name", "value",
			"expiry_date", "version", "secure", "comment", "comment_url" };

	/* create table queries */
	private static final String SQL_CREATE_TABLE = "CREATE TABLE ";
	private static final String SQL_QUERY_CREATE_TABLE_DBINFO = SQL_CREATE_TABLE + TABLE_DBINFO + " (" + COLS_DBINFO[0]
			+ " TEXT PRIMARY KEY, " + COLS_DBINFO[1] + " TEXT, " + COLS_DBINFO[2] + " INTEGER)";
	private static final String SQL_QUERY_CREATE_TABLE_OFFLINE_SEARCH = SQL_CREATE_TABLE + TABLE_OFFLINE_SEARCH + " ("
			+ COLS_OFFLINE_SEARCH[0] + " INTEGER PRIMARY KEY, " + COLS_OFFLINE_SEARCH[1] + " INTEGER NOT NULL, "
			+ COLS_OFFLINE_SEARCH[2] + " TEXT, " + COLS_OFFLINE_SEARCH[3] + " TEXT, " + COLS_OFFLINE_SEARCH[4]
			+ " INTEGER NOT NULL)";
	private static final String SQL_QUERY_CREATE_TABLE_OFFLINE_RESULT = SQL_CREATE_TABLE + TABLE_OFFLINE_RESULTS
			+ " ( " + COLS_OFFLINE_RESULT[0] + " INTEGER NOT NULL, " + COLS_OFFLINE_RESULT[1] + " INTEGER NOT NULL)";
	private static final String SQL_QUERY_CREATE_TABLE_ONLINE_SEARCH = SQL_CREATE_TABLE + TABLE_ONLINE_SEARCH + " ("
			+ COLS_ONLINE_SEARCH[0] + " INTEGER PRIMARY KEY, " + COLS_ONLINE_SEARCH[1] + " INTEGER NOT NULL, "
			+ COLS_ONLINE_SEARCH[2] + " TEXT NOT NULL, " + COLS_ONLINE_SEARCH[3] + " INTEGER NOT NULL, "
			+ COLS_ONLINE_SEARCH[4] + " TEXT NOT NULL, " + COLS_ONLINE_SEARCH[5] + " TEXT NOT NULL, "
			+ COLS_ONLINE_SEARCH[6] + " TEXT NOT NULL, " + COLS_ONLINE_SEARCH[7] + " TEXT NOT NULL)";
	private static final String SQL_QUERY_CREATE_TABLE_ONLINE_COOKIES = SQL_CREATE_TABLE + TABLE_ONLINE_COOKIES + " ("
			+ COLS_ONLINE_COOKIES[0] + " TEXT NOT NULL, " + COLS_ONLINE_COOKIES[1] + " TEXT, " + COLS_ONLINE_COOKIES[2]
			+ " TEXT, " + COLS_ONLINE_COOKIES[3] + " TEXT NOT NULL, " + COLS_ONLINE_COOKIES[4] + " TEXT, "
			+ COLS_ONLINE_COOKIES[5] + " INTEGER NOT NULL, " + COLS_ONLINE_COOKIES[6] + " INTEGER, "
			+ COLS_ONLINE_COOKIES[7] + " INTEGER, " + COLS_ONLINE_COOKIES[8] + " TEXT, " + COLS_ONLINE_COOKIES[9]
			+ " TEXT)";

	/* keys */
	static final String KEY_DB_DICT_VERSION = "dict_version";
	static final String KEY_DB_DICT_CREATION_TIME = "dict_creation_time";
	/* values */
	static final int VALUE_DB_VERSION = 1;

	/* select queries */
	private static final String QUERY_OFFLINE_SEARCH = COLS_OFFLINE_SEARCH[0] + "=?";
	private static final String QUERY_OFFLINE_RESULT = COLS_OFFLINE_RESULT[0] + "=?";
	private static final String QUERY_ONLINE_SEARCH = COLS_ONLINE_SEARCH[0] + "=?";
	private static final String QUERY_ONLINE_COOKIES = COLS_ONLINE_COOKIES[0] + "=?";
	private static final String QUERY_ONLINE_COOKIES_EXPIRY = COLS_ONLINE_COOKIES[5] + "<=?";

	/* instance variables */
	private SQLiteDatabase database;
	private static HistoryDB instance = null;

	private HistoryDB() {
		this.createDatabaseIfNeeded();
	}

	@Override
	protected void finalize() throws Throwable {
		super.finalize();
		if (null != this.database && this.database.isOpen()) {
			// be sure to close the database
			this.database.close();
			this.database = null;
		}
	}

	public static HistoryDB getHistoryDB() {
		if (null == HistoryDB.instance) {
			HistoryDB.instance = new HistoryDB();
		}

		return HistoryDB.instance;
	}

	protected static void cleanDB() {
		if (null != HistoryDB.instance) {
			HistoryDB.instance.database.close();
			HistoryDB.instance.database = null;
			HistoryDB.instance = null;
		}
	}

	public ArrayList<SearchResult> getAllOfflineResult() {
		final SQLiteDatabase db = this.database;

		Cursor cursor = db.query(TABLE_OFFLINE_SEARCH, COLS_OFFLINE_SEARCH, null, null, null, null, null);
		final int numSearchs = cursor.getCount();

		final ArrayList<SearchResult> alSR = new ArrayList<SearchResult>(numSearchs);
		for (int i = 0; i < numSearchs; i++) {
			if (cursor.moveToNext()) {
				int colIdx = 1;
				long date = cursor.getLong(colIdx++);
				String term = cursor.getString(colIdx++);
				Language srcLang = Language.MAP_ISO3.get(cursor.getString(colIdx++));
				int numResults = cursor.getInt(colIdx++);

				OfflineResult offlineResult = new OfflineResult(new Date(date), term, srcLang, numResults);
				alSR.add(offlineResult);
			}
		}
		cursor.close();

		final String[] selectionArgs = new String[1];
		for (int i = 0; i < numSearchs; i++) {
			final SearchResult searchResult = alSR.get(i);
			final OfflineResult offlineResult = (OfflineResult) searchResult;

			selectionArgs[0] = Long.toString(searchResult.getId());
			cursor = db.query(TABLE_OFFLINE_RESULTS, COLS_OFFLINE_RESULT, QUERY_OFFLINE_RESULT, selectionArgs, null,
					null, null);
			final int numResults = cursor.getCount();
			for (int r = 0; r < numResults; r++) {
				if (cursor.moveToNext()) {
					offlineResult.dictionaryIds.add(cursor.getInt(1));
				}
			}
			cursor.close();
		}

		Collections.sort(alSR);

		return alSR;

	}

	public void addOffileResult(OfflineResult offlineResult) {

		final SQLiteDatabase db = this.database;

		db.beginTransaction();
		try {
			final ArrayList<Integer> resultIds = offlineResult.dictionaryIds;
			final int numResult = resultIds.size();
			this.deleteOfflineResult(offlineResult);
			final long id = offlineResult.getId();
			ContentValues search = new ContentValues(COLS_OFFLINE_SEARCH.length);

			search.put(COLS_OFFLINE_SEARCH[0], id);
			search.put(COLS_OFFLINE_SEARCH[1], offlineResult.date.getTime());
			search.put(COLS_OFFLINE_SEARCH[2], offlineResult.searchTerm);
			search.put(COLS_OFFLINE_SEARCH[3], offlineResult.sourceLang.langISO3);
			search.put(COLS_OFFLINE_SEARCH[4], numResult);

			long insertCode = db.insert(TABLE_OFFLINE_SEARCH, null, search);

			if (-1 == insertCode) {
				// silently fail
				return;
			}

			ContentValues result = new ContentValues(COLS_OFFLINE_RESULT.length);

			for (int i = 0; i < numResult; i++) {
				result.clear();
				result.put(COLS_OFFLINE_RESULT[0], id);
				result.put(COLS_OFFLINE_RESULT[1], resultIds.get(i));
				db.insert(TABLE_OFFLINE_RESULTS, null, result);
			}

			db.setTransactionSuccessful();
		} catch (SQLiteException sle) {
			Log.e(TAG, sle.getMessage(), sle);
		} finally {
			db.endTransaction();
		}
	}

	/**
	 * Warn: Not in a transaction.
	 * 
	 * @param date
	 *            to delete
	 */
	private void deleteOfflineResult(OfflineResult offlineResult) {
		final SQLiteDatabase db = this.database;

		final String[] selectionArgs = new String[1];
		selectionArgs[0] = Long.toString(offlineResult.getId());

		try {
			db.delete(TABLE_OFFLINE_SEARCH, QUERY_OFFLINE_SEARCH, selectionArgs);
			db.delete(TABLE_OFFLINE_RESULTS, QUERY_OFFLINE_RESULT, selectionArgs);
		} catch (SQLiteException sle) {
			Log.e(TAG, sle.getMessage(), sle);
		}
	}

	public void deleteCookies() {
		final SQLiteDatabase db = this.database;
		db.beginTransaction();
		try {
			db.delete(TABLE_ONLINE_COOKIES, null, null);
			db.setTransactionSuccessful();
		} catch (SQLiteException sle) {
			Log.e(TAG, sle.getMessage(), sle);
		} finally {
			db.endTransaction();
		}
	}

	public CookieStore getCookies(String serviceName) {
		final BasicCookieStore cookieStore = new BasicCookieStore();

		final SQLiteDatabase db = this.database;

		String where = null;
		String[] selectionArgs = null;

		if (null != serviceName) {
			where = QUERY_ONLINE_COOKIES;
			selectionArgs = new String[1];
			selectionArgs[0] = serviceName;
		}

		try {
			final Cursor cursor = db.query(TABLE_ONLINE_COOKIES, COLS_ONLINE_COOKIES, where, selectionArgs, null, null,
					null, null);

			while (cursor.moveToNext()) {

				String domain = cursor.getString(1);
				String path = cursor.getString(2);
				String name = cursor.getString(3);
				String value = cursor.getString(4);
				Date expiryDate = new Date(cursor.getLong(5));
				int version = 0;
				if (!cursor.isNull(6)) {
					version = cursor.getInt(6);
				}
				boolean secure = false;
				if (!cursor.isNull(7)) {
					secure = cursor.getInt(7) == 0 ? false : true;
				}
				String comment = cursor.getString(8);
				String commentURL = cursor.getString(9);

				BasicClientCookie2 cookie = new BasicClientCookie2(name, value);
				cookie.setDomain(domain);
				cookie.setPath(path);
				cookie.setExpiryDate(expiryDate);
				cookie.setVersion(version);
				cookie.setSecure(secure);
				cookie.setComment(comment);
				cookie.setCommentURL(commentURL);

				cookieStore.addCookie(cookie);
			}
			cursor.close();
		} catch (SQLiteException sle) {
			Log.e(TAG, sle.getMessage(), sle);
		}
		return cookieStore;
	}

	public void addCookies(String serviceName, CookieStore cookies) {
		final SQLiteDatabase db = this.database;
		db.beginTransaction();
		try {
			for (Cookie cookie : cookies.getCookies()) {
				if (cookie.isPersistent()) {
					ContentValues values = new ContentValues();
					values.put(COLS_ONLINE_COOKIES[0], serviceName);
					values.put(COLS_ONLINE_COOKIES[1], cookie.getDomain());
					values.put(COLS_ONLINE_COOKIES[2], cookie.getPath());
					values.put(COLS_ONLINE_COOKIES[3], cookie.getName());
					values.put(COLS_ONLINE_COOKIES[4], cookie.getValue());
					if (null != cookie.getExpiryDate())
						values.put(COLS_ONLINE_COOKIES[5], cookie.getExpiryDate().getTime());
					values.put(COLS_ONLINE_COOKIES[6], cookie.getVersion());
					values.put(COLS_ONLINE_COOKIES[7], cookie.isSecure() ? 1 : 0);
					values.put(COLS_ONLINE_COOKIES[8], cookie.getComment());
					values.put(COLS_ONLINE_COOKIES[9], cookie.getCommentURL());

					db.insert(TABLE_ONLINE_COOKIES, null, values);
				}
			}
			db.setTransactionSuccessful();
		} catch (SQLiteException sle) {
			Log.e(TAG, sle.getMessage(), sle);
		} finally {
			db.endTransaction();
		}
	}

	public void addOnlineResult(OnlineResult onlineResult) {
		final SQLiteDatabase db = this.database;

		db.beginTransaction();
		try {

			this.deleteOnlineResults(onlineResult);

			ContentValues values = new ContentValues();

			int idx = 0;
			values.put(COLS_ONLINE_SEARCH[idx++], onlineResult.getId());
			values.put(COLS_ONLINE_SEARCH[idx++], onlineResult.date.getTime());
			values.put(COLS_ONLINE_SEARCH[idx++], onlineResult.searchTerm);
			values.put(COLS_ONLINE_SEARCH[idx++], onlineResult.numResults);
			values.put(COLS_ONLINE_SEARCH[idx++], onlineResult.sourceLang.langISO3);
			values.put(COLS_ONLINE_SEARCH[idx++], onlineResult.destLang.langISO3);
			values.put(COLS_ONLINE_SEARCH[idx++], onlineResult.service);
			values.put(COLS_ONLINE_SEARCH[idx++], onlineResult.html);

			db.insert(TABLE_ONLINE_SEARCH, null, values);
			db.setTransactionSuccessful();
		} catch (SQLiteException sle) {
			Log.e(TAG, sle.getMessage(), sle);
		} finally {
			db.endTransaction();
		}
	}

	public ArrayList<SearchResult> getAllOnlineResults() {
		final SQLiteDatabase db = this.database;

		final Cursor cursor = db.query(TABLE_ONLINE_SEARCH, COLS_ONLINE_SEARCH, null, null, null, null, null);
		final ArrayList<SearchResult> list = new ArrayList<SearchResult>(cursor.getCount());

		while (cursor.moveToNext()) {
			OnlineResult result = new OnlineResult(new Date(cursor.getLong(1)), cursor.getString(2), cursor.getInt(3),
					Language.MAP_ISO3.get(cursor.getString(4)), Language.MAP_ISO3.get(cursor.getString(5)),
					cursor.getString(6), cursor.getString(7));
			list.add(result);
		}
		cursor.close();

		Collections.sort(list);

		return list;
	}

	/**
	 * Warn: Not in a transaction.
	 * 
	 * @param date
	 *            of the search
	 */
	private void deleteOnlineResults(OnlineResult onlineResult) {
		final String[] selectionArgs = new String[1];
		selectionArgs[0] = Long.toString(onlineResult.getId());

		this.database.delete(TABLE_ONLINE_SEARCH, QUERY_ONLINE_SEARCH, selectionArgs);
	}

	private void deleteExpiredCookies() {
		final String[] selectionArgs = new String[1];
		selectionArgs[0] = Long.toString(System.currentTimeMillis());
		this.database.delete(TABLE_ONLINE_COOKIES, QUERY_ONLINE_COOKIES_EXPIRY, selectionArgs);
	}

	private void createDatabaseIfNeeded() {
		final File dbFile = new File(Common.cachePath, DB_FILENAME);
		final int maxSearch = Preferences.getInt(R.string.prefNameHistoryDictionaryMax);

		if (0 == maxSearch && dbFile.isFile()) {
			// Delete the history file, if the user want to
			dbFile.delete();
		}

		final InfosDB dictMainDB = CommonDB.getDBInfos(DatabaseType.dictonary, Language.JAPANESE);

		if (null == dictMainDB || 0 == maxSearch) {
			// Create a memory only database
			this.database = SQLiteDatabase.create(null);
			this.initDB(0);
		} else if (!dbFile.isFile()) {
			try {
				this.database = SQLiteDatabase.openDatabase(dbFile.getAbsolutePath(), null,
						SQLiteDatabase.CREATE_IF_NECESSARY | SQLiteDatabase.OPEN_READWRITE);
			} catch (SQLiteException sle) {
				// Create a memory only database
				this.database = SQLiteDatabase.create(null);
			}

			this.initDB(dictMainDB.creationDate.getTime());
		} else {
			this.database = SQLiteDatabase.openDatabase(dbFile.getAbsolutePath(), null, SQLiteDatabase.OPEN_READWRITE);
			this.checkDBVersion(dictMainDB.creationDate.getTime());

			final SQLiteDatabase db = this.database;
			boolean needVacuum = false;
			boolean haveException = false;

			db.beginTransaction();
			try {

				// If there is more search than the max, delete them
				// may be optimize, but is it usefull ?
				final ArrayList<SearchResult> alSR = this.getAllOfflineResult();
				final int numOfflineSearch = alSR.size();
				if (numOfflineSearch > maxSearch) {
					needVacuum = true;
					for (int i = maxSearch; i < numOfflineSearch; i++) {
						this.deleteOfflineResult((OfflineResult) alSR.get(i));
					}
				}

				final ArrayList<SearchResult> alOS = this.getAllOnlineResults();
				final int numOnlineSearch = alOS.size();
				if (numOnlineSearch > maxSearch) {
					needVacuum = true;

					for (int i = maxSearch; i < numOnlineSearch; i++) {
						this.deleteOnlineResults((OnlineResult) alOS.get(i));
					}
				}

				this.deleteExpiredCookies();

				if (needVacuum) {
					db.execSQL("VACUUM");
				}
				db.setTransactionSuccessful();
			} catch (SQLiteException sle) {
				Log.e(TAG, sle.getMessage(), sle);
				haveException = true;
			} finally {
				db.endTransaction();
			}

			if (haveException) {
				Log.i(TAG, "delete and rebuild " + dbFile.getName());
				this.database.close();
				this.database = null;
				dbFile.delete();
				this.createDatabaseIfNeeded();
			}
		}
	}

	private void initDB(long dictDBCreationTime) {
		this.database.execSQL(SQL_QUERY_CREATE_TABLE_DBINFO);
		this.database.execSQL(SQL_QUERY_CREATE_TABLE_OFFLINE_SEARCH);
		this.database.execSQL(SQL_QUERY_CREATE_TABLE_OFFLINE_RESULT);
		this.database.execSQL(SQL_QUERY_CREATE_TABLE_ONLINE_SEARCH);
		this.database.execSQL(SQL_QUERY_CREATE_TABLE_ONLINE_COOKIES);

		this.database.beginTransaction();
		try {

			ContentValues dbInfos = new ContentValues(2);

			dbInfos.put(COLS_DBINFO[0], KEY_DB_VERSION);
			dbInfos.put(COLS_DBINFO[2], VALUE_DB_VERSION);
			this.database.insert(TABLE_DBINFO, null, dbInfos);
			dbInfos.clear();

			dbInfos.put(COLS_DBINFO[0], KEY_DB_CREATION_TIME);
			dbInfos.put(COLS_DBINFO[2], Calendar.getInstance().getTimeInMillis());
			this.database.insert(TABLE_DBINFO, null, dbInfos);
			dbInfos.clear();

			dbInfos.put(COLS_DBINFO[0], KEY_DB_DICT_VERSION);
			dbInfos.put(COLS_DBINFO[2], DictionaryDB.VALUE_DB_VERSION);
			this.database.insert(TABLE_DBINFO, null, dbInfos);
			dbInfos.clear();

			dbInfos.put(COLS_DBINFO[0], KEY_DB_DICT_CREATION_TIME);
			dbInfos.put(COLS_DBINFO[2], dictDBCreationTime);
			this.database.insert(TABLE_DBINFO, null, dbInfos);
			dbInfos.clear();

			this.database.setTransactionSuccessful();
		} catch (SQLiteException sle) {
			Log.e(TAG, sle.getMessage(), sle);
		} finally {
			this.database.endTransaction();
		}
	}

	private void checkDBVersion(long dictDBCreationTime) {
		// int dbVersion = 0;
		int dbDictVersion = 0;
		long dbDictCreationTime = 0;

		final Cursor cursor = this.database.query(CommonDB.TABLE_DBINFO, CommonDB.COLS_DBINFO, null, null, null, null,
				null);
		if (cursor.moveToFirst()) {
			final int idxColKey = cursor.getColumnIndex(CommonDB.COLS_DBINFO[0]);
			// final int idxColValueStr =
			// cursor.getColumnIndex(CommonDB.COLS_DBINFO[1]);
			final int idxColValueInt = cursor.getColumnIndex(CommonDB.COLS_DBINFO[2]);
			do {
				String key = cursor.getString(idxColKey);
				// if (CommonDB.KEY_DB_VERSION.equals(key)) {
				// dbVersion = cursor.getInt(idxColValueInt);
				// } else
				if (HistoryDB.KEY_DB_DICT_CREATION_TIME.equals(key)) {
					dbDictCreationTime = cursor.getLong(idxColValueInt);
				} else if (HistoryDB.KEY_DB_DICT_VERSION.equals(key)) {
					dbDictVersion = cursor.getInt(idxColValueInt);
				}
			} while (cursor.moveToNext());
		}
		cursor.close();

		// Format change
		// switch (dbVersion) {
		// case 1:
		// // future
		// }

		if (DictionaryDB.VALUE_DB_VERSION != dbDictVersion || dbDictCreationTime != dictDBCreationTime) {
			// Dictionary DB change

			// Clear the search results
			this.database.beginTransaction();
			try {
				// delete results, dictionary ids may have change
				this.database.delete(TABLE_OFFLINE_RESULTS, null, null);

				// Update dict db infos
				final String[] selectionArgs = new String[1];
				selectionArgs[0] = KEY_DB_DICT_VERSION;
				this.database.delete(TABLE_DBINFO, COLS_DBINFO[0] + "=?", selectionArgs);
				selectionArgs[0] = KEY_DB_DICT_CREATION_TIME;
				this.database.delete(TABLE_DBINFO, COLS_DBINFO[0] + "=?", selectionArgs);

				ContentValues dbInfos = new ContentValues(2);
				dbInfos.put(COLS_DBINFO[0], KEY_DB_DICT_VERSION);
				dbInfos.put(COLS_DBINFO[2], DictionaryDB.VALUE_DB_VERSION);
				this.database.insert(TABLE_DBINFO, null, dbInfos);
				dbInfos.clear();

				dbInfos.put(COLS_DBINFO[0], KEY_DB_DICT_CREATION_TIME);
				dbInfos.put(COLS_DBINFO[2], dictDBCreationTime);
				this.database.insert(TABLE_DBINFO, null, dbInfos);
				dbInfos.clear();

				this.database.setTransactionSuccessful();

			} catch (SQLiteException sle) {
				Log.e(TAG, sle.getMessage(), sle);
			} finally {
				this.database.endTransaction();
			}
		}
	}
}
