/*
 * Copyright (C) 2012 Yvon TANGUY
 * 
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.db;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Date;
import java.util.Locale;
import java.util.regex.Matcher;

import org.vono.narau.R;
import org.vono.narau.common.Common;
import org.vono.narau.utils.Language;

import android.database.sqlite.SQLiteDatabase;
import android.util.Log;
import android.widget.CheckBox;

/**
 * DB informations: type, lang, name etc.
 * <p>
 * gz_filename db_creation_time db_size gz_size gz_md5 download_url
 * another_download_url...
 * </p>
 * <p>
 * Please note that this class do not contains the db version, since there is
 * only instances of database with a good version.
 * </p>
 */
public class InfosDB implements Comparable<InfosDB> {
	private static final String TAG = InfosDB.class.getSimpleName();

	private static final String WARN_LOG_DB_VERSION = "Wrong db version for %1$s %2$s: %3$d expected %4$d";

	public enum Action {
		none, download, delete
	}

	// DB Info
	public final DatabaseType type;
	public final Language language;
	public final String dbName;
	/**
	 * R.string.dbDictionary / R.string.dbKanjiKana
	 */
	public final String userNameDBType;
	public final File file;
	final SQLiteDatabase db;
	Date creationDate;

	// Download infos:
	public long gzSize;
	public long dbSize;
	public Date downloadCreationDate;
	public String md5;
	public URL[] url;
	// Download UI:
	public CheckBox checkBox;
	public Action action;

	private InfosDB(DatabaseType type, Language language) {
		this.type = type;
		this.language = language;

		if (this.type == DatabaseType.dictonary) {
			this.userNameDBType = Common.getString(R.string.dbDictionary).intern();
		} else {
			this.userNameDBType = Common.getString(R.string.dbKanjiKana).intern();
		}

		this.dbName = CommonDB.getDBFilename(this.type, language).intern();
		this.dbSize = 0;
		this.gzSize = 0;
		this.creationDate = null;
		this.downloadCreationDate = null;
		this.checkBox = null;

		this.file = new File(Common.dataPath, this.dbName);
		this.db = null;
		this.action = Action.none;
	}

	InfosDB(DatabaseType type, Language language, SQLiteDatabase db) {
		// The duplication of code is mandatory, because of final class members.
		this.type = type;
		this.language = language;

		if (this.type == DatabaseType.dictonary) {
			this.userNameDBType = Common.getString(R.string.dbDictionary).intern();
		} else {
			this.userNameDBType = Common.getString(R.string.dbKanjiKana).intern();
		}

		this.dbName = CommonDB.getDBFilename(this.type, language).intern();
		this.dbSize = 0;
		this.gzSize = 0;
		this.creationDate = null;
		this.checkBox = null;

		this.file = new File(Common.dataPath, this.dbName);
		this.db = db;
		this.action = Action.none;
	}

	@Override
	protected void finalize() throws Throwable {
		super.finalize();
		if (null != this.db && this.db.isOpen()) {
			this.db.close();
		}
	}

	void setCreationTime(long milliseconds) {
		// Unix file time: precision up to 1 second
		// Java Date: miliseconds
		this.creationDate = new Date(milliseconds);
	}

	private void setDownloadCreationTime(long milliseconds) {
		this.downloadCreationDate = new Date(milliseconds);
	}

	public boolean isSelected() {
		if (null != this.checkBox) {
			return (this.checkBox.isChecked() && this.checkBox.isEnabled());
		}
		return false;
	}

	public boolean haveLatest() {
		// Current creationDate set (not null)
		// Current creationDate before downloadCreationDate
		if (null != this.creationDate) {
			return (!this.creationDate.before(this.downloadCreationDate));
		}
		return false;
	}

	public boolean exists() {
		return this.file.isFile();
	}

	public boolean isMain() {
		return Language.JAPANESE.equals(this.language);
	}

	@Override
	public String toString() {
		return this.dbName;
	}

	public int compareTo(InfosDB another) {
		int diff = this.type.getValue() - another.type.getValue();
		if (diff == 0) {
			// Put the main DB language first
			if (this.language.equals(Language.JAPANESE)) {
				diff = -1;
			} else if (another.language.equals(Language.JAPANESE)) {
				diff = 1;
			} else {
				diff = this.language.compareTo(another.language);
			}
		}
		return diff;
	}

	@Override
	public boolean equals(Object o) {
		if (o instanceof InfosDB) {
			InfosDB other = (InfosDB) o;
			return this.dbName == other.dbName;
		}
		return false;
	}

	@Override
	public int hashCode() {
		return this.dbName.hashCode();
	}

	/**
	 * Parse a single line in narau_db_infos.txt file<br/>
	 * 
	 * <pre>
	 * db_filename db_creation_time db_size gz_size gz_md5 download_url another_download_url...
	 * </pre>
	 * 
	 * @param line
	 *            the line to parse
	 * @return null if the line is invalid
	 * @throws NumberFormatException
	 *             If error converting string to number
	 */
	public static InfosDB getDBInfoFromLine(String line) {
		// Parse the line
		final String[] infos = line.split(" +");
		final int numInfos = infos.length;
		if (numInfos < 6) {
			Log.e(TAG, "Not enough items on the line: " + line);
			return null;
		}
		int pos = 0;
		final String dbName = infos[pos++].trim();
		DatabaseType type = DatabaseType.dictonary;
		Matcher matcher = CommonDB.PATTERN_DICT.matcher(dbName);
		if (!matcher.matches()) {
			matcher = CommonDB.PATTERN_KANJI.matcher(dbName);
			if (matcher.matches()) {
				type = DatabaseType.kanji;
			} else {
				Log.e(TAG, dbName + ": do not match any db name pattern");
				return null;
			}
		}

		final String lang = matcher.group(1);
		final int dbVersion = Integer.parseInt(matcher.group(2));

		// check version, must be good, or there is an error in
		// narau_db_infos.txt
		if (type == DatabaseType.dictonary && DictionaryDB.VALUE_DB_VERSION != dbVersion) {
			Log.w(TAG, String.format(WARN_LOG_DB_VERSION, type.name(), lang, dbVersion, DictionaryDB.VALUE_DB_VERSION));
			return null;
		} else if (type == DatabaseType.kanji && KanjiDB.VALUE_DB_VERSION != dbVersion) {
			Log.w(TAG, String.format(WARN_LOG_DB_VERSION, type.name(), lang, dbVersion, KanjiDB.VALUE_DB_VERSION));
			return null;
		}

		final long dbCreationTime = Long.parseLong(infos[pos++]);
		final long dbSize = Long.parseLong(infos[pos++]);
		final long gzSize = Long.parseLong(infos[pos++]);
		final String md5 = infos[pos++].toLowerCase(Locale.US);
		// download URLs:
		final ArrayList<URL> listDownloadURL = new ArrayList<URL>();
		for (int i = pos; i < numInfos; i++) {
			try {
				final URL url = new URL(infos[i]);
				listDownloadURL.add(url);
			} catch (MalformedURLException mue) {
				Log.w(TAG, "Bad URL " + infos[i], mue);
			}
		}
		if (listDownloadURL.isEmpty()) {
			Log.w(TAG, "No valid url found in " + line);
			return null;
		}

		Language language = Language.MAP_ISO3.get(lang);

		// Search for an existing dbInfo which is the same as this one
		InfosDB infosDB = CommonDB.getDBInfos(type, language);

		if (null == infosDB) {
			// Not found, create a new one
			infosDB = new InfosDB(type, language);
			CommonDB.listDBInfos.add(infosDB);
		}

		// Set the download values
		infosDB.setDownloadCreationTime(dbCreationTime);
		infosDB.dbSize = dbSize;
		infosDB.gzSize = gzSize;
		infosDB.md5 = md5;
		infosDB.url = listDownloadURL.toArray(new URL[0]);

		return infosDB;
	}
}
