/*
 *  Copyright (C) 2012 Yvon TANGUY
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.dictionary;

import java.util.ArrayList;

import org.vono.narau.R;
import org.vono.narau.common.dictionary.EntryArrayAdapter;
import org.vono.narau.common.dictionary.EntryListFragment;
import org.vono.narau.db.CommonDB;
import org.vono.narau.db.DatabaseType;
import org.vono.narau.models.dictionary.Entry;
import org.vono.narau.models.history.OfflineResult;
import org.vono.narau.models.history.SearchResult;
import org.vono.narau.preferences.Preferences;
import org.vono.narau.utils.Language;

import android.app.AlertDialog;
import android.content.Context;
import android.content.DialogInterface;
import android.content.DialogInterface.OnClickListener;
import android.content.res.Configuration;
import android.os.Bundle;
import android.os.Handler;
import android.os.Message;
import android.view.Menu;
import android.view.MenuInflater;
import android.view.MenuItem;
import android.view.View;
import android.view.inputmethod.InputMethodManager;
import android.widget.Toast;

public class DictionaryOfflineActivity extends DictionaryBase implements OnClickListener {
	// private static final String TAG =
	// DictionaryOfflineActivity.class.getSimpleName();

	private static final String[][] TEXT_REPLACE = { { "'", "''" }, { ";", "" }, { "\"", "" }, { "/*", "" },
			{ "*/", "" }, { "--", "" } };

	private static final String AUTO = "auto";

	/**
	 * Tag messages for buttonSearch to be able to do the search or stop the loading thread.
	 */
	private static final String BUTTON_DO_SEARCH = "DS";
	private static final String BUTTON_STOP_LOADING = "SL";

	private static DictionaryThread thread = null;

	private EntryListFragment entryListFragment;

	private DictionaryHandler handler;

	/**
	 * dummy language.
	 */
	private Language languageAuto;
	/**
	 * Dictionary language & languageAuto
	 */
	private ArrayList<Language> languages;

	public DictionaryOfflineActivity() {
		super(DictionaryType.offline);
		this.entryListFragment = null;
		this.languages = null;
	}

	public static boolean haveDatabase() {
		// Method used by Narau.showMissingDatabaseNotification()
		return CommonDB.haveDB(DatabaseType.dictonary);
	}

	/**
	 * Called when the activity is first created.
	 */
	@Override
	public void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);

		this.handler = new DictionaryHandler(this);

		final ArrayList<Language> languagesDictionary = Preferences.getLangsPriorities(DatabaseType.dictonary);
		final int numLangs = languagesDictionary.size();
		final ArrayList<Language> languagesDisplay = new ArrayList<Language>(numLangs + 2);
		this.languageAuto = new Language(AUTO, AUTO, AUTO, super.getString(R.string.dictionaryOfflineAutoLang));
		languagesDisplay.add(this.languageAuto);
		languagesDisplay.add(Language.JAPANESE);
		for (int i = 0; i < numLangs; i++) {
			languagesDisplay.add(languagesDictionary.get(i));
		}
		this.languages = languagesDisplay;

		super.setContentView(R.layout.dictionary_offline);

		this.entryListFragment = (EntryListFragment) super.getSupportFragmentManager().findFragmentById(
				R.id.dictionaryFragmentEntryList);

		super.buttonSearch.setOnLongClickListener(this);
	}

	@Override
	protected void onPostCreate(Bundle savedInstanceState) {
		Bundle bundle = savedInstanceState;
		if (null == bundle) {
			bundle = super.getIntent().getExtras();

			if (null == bundle) {
				// Empty bundle
				bundle = new Bundle();
			}
		}

		super.onPostCreate(bundle);

		// Default values
		if (null == super.languageSource) {
			super.languageSource = this.languageAuto;
		}
		super.languageDest = this.languageAuto;

		if (bundle.containsKey(BUTTON_DO_SEARCH)) {
			super.buttonSearch.setTag(bundle.getString(BUTTON_DO_SEARCH));
		} else {
			super.buttonSearch.setTag(BUTTON_DO_SEARCH);
		}
	}

	@Override
	protected void onSaveInstanceState(Bundle outState) {
		super.onSaveInstanceState(outState);
		outState.putString(BUTTON_DO_SEARCH, super.buttonSearch.getTag().toString());
	}

	@Override
	protected void onDestroy() {
		super.onDestroy();

		if (super.isFinishing()) {

			this.stopDictionaryThread(false);
			this.handler = null;

			this.entryListFragment.clear();
			this.entryListFragment = null;

			System.gc();
			System.runFinalization();
		}
	}

	@Override
	protected void onPause() {
		super.onPause();

		if (null != DictionaryOfflineActivity.thread) {
			DictionaryOfflineActivity.thread.pause();
		}
	}

	@Override
	protected void onResume() {
		super.onResume();

		if (null != DictionaryOfflineActivity.thread) {
			this.setWidgetEnabled(false);
			DictionaryOfflineActivity.thread.setStopWaitingActivity(this.handler);
		}

		// Put here because when the activity is recreated, the search button
		// may be disabled, event if we may stop the loading thread.
		// This function is called on the first init anyway.
		// It would be more logical to have it in onPostCreate() though.
		this.updateButtonSearchText();
	}

	@Override
	public void onClickSearch(View button) {

		if (BUTTON_DO_SEARCH.equals(super.buttonSearch.getTag())) {
			final String searchText = DictionaryOfflineActivity
					.secureTextMatch(this.editTextInput.getText().toString());

			if (!searchText.isEmpty()) {

				// Hide the keyboard
				if (super.getResources().getConfiguration().keyboardHidden == Configuration.KEYBOARDHIDDEN_NO) {
					InputMethodManager imm = (InputMethodManager) getSystemService(Context.INPUT_METHOD_SERVICE);
					imm.hideSoftInputFromWindow(this.editTextInput.getWindowToken(), 0);
				}

				this.search(searchText, button != null);
			}
		} else {
			this.stopDictionaryThread(true);
			super.buttonSearch.setTag(BUTTON_DO_SEARCH);
			this.updateButtonSearchText();
		}
	}

	public void onHistorySearchSelected(SearchResult sr) {
		final OfflineResult offlineResult = (OfflineResult) sr;

		super.languageSource = sr.sourceLang;
		super.languageDest = sr.destLang;

		if (sr.numResults != offlineResult.dictionaryIds.size()) {
			super.showProgressDialog();
		}

		this.updateButtonSearchText();
		this.setWidgetEnabled(false);
		this.entryListFragment.clear();
		this.editTextInput.setText(offlineResult.searchTerm);

		this.stopDictionaryThread(true);
		DictionaryOfflineActivity.thread = new DictionaryThread(Thread.currentThread(), this.handler, offlineResult);
		DictionaryOfflineActivity.thread.start();
	}

	public boolean onLongClick(View v) {
		AlertDialog.Builder builder = new AlertDialog.Builder(this);

		int currentLang = this.languages.indexOf(super.languageSource);

		builder.setSingleChoiceItems(this.languages.toArray(new Language[0]), currentLang, this);

		builder.show();
		return true;
	}

	/**
	 * Manage the alert dialog validation, to select the search database.
	 * <p/>
	 */
	public void onClick(DialogInterface dialog, int which) {
		dialog.dismiss();

		super.languageSource = this.languages.get(which);
		this.updateButtonSearchText();

		if (super.languageSource.equals(this.languageAuto)) {
			super.languageDest = this.languageAuto;
		} else if (Language.JAPANESE.equals(this.languageSource)) {
			super.languageDest = this.languageAuto;
		} else {
			super.languageDest = Language.JAPANESE;
		}
		Preferences.setString(R.string.prefNameDictionaryOffLineLangPair, super.languageSource.langISO3 + LANG_PAIR_SEP
				+ super.languageDest.langISO3);
	}

	@Override
	public void onClickReset(View button) {
		this.stopDictionaryThread(true);
		super.onClickReset(button);
	}

	/**
	 * Change the text size.
	 * 
	 * @param buttonZoom
	 */
	@Override
	public void onClickZoom(final View buttonZoom) {
		super.onClickZoom(buttonZoom);
		((EntryArrayAdapter) this.entryListFragment.getListAdapter()).notifyDataSetChanged();
	}

	@Override
	public boolean onCreateOptionsMenu(Menu menu) {
		MenuInflater inflater = getMenuInflater();
		inflater.inflate(R.menu.dictionary_offline, menu);

		MenuItem item = menu.findItem(R.id.menuDictionaryWordMatch);
		updateMenuItemWordMatch(item, Preferences.getBoolean(R.string.prefNameDictionaryWordMatch));
		return true;
	}

	@Override
	public boolean onOptionsItemSelected(MenuItem item) {
		switch (item.getItemId()) {
		case R.id.menuDictionaryLoadAll:
			this.onClickSearch(null);
			break;
		case R.id.menuDictionaryWordMatch:
			boolean wordMatch = !Preferences.getBoolean(R.string.prefNameDictionaryWordMatch);
			Preferences.setBoolean(R.string.prefNameDictionaryWordMatch, wordMatch);
			updateMenuItemWordMatch(item, wordMatch);
			break;
		case R.id.menuDictionarySelectLang:
			this.onLongClick(null);
			break;
		default:
			return super.onOptionsItemSelected(item);
		}

		return true;
	}

	private static void updateMenuItemWordMatch(MenuItem item, boolean wordMatch) {
		if (wordMatch) {
			item.setTitle(R.string.dictionaryOfflineWordExtendedMatch);
			item.setIcon(R.drawable.ic_menu_dictionary_wordextended);
		} else {
			item.setTitle(R.string.dictionaryOfflineWordExactMatch);
			item.setIcon(R.drawable.ic_menu_dictionary_wordmatch);
		}
	}

	@Override
	protected void updateButtonSearchText() {

		if (BUTTON_DO_SEARCH.equals(super.buttonSearch.getTag())) {
			if (this.languageAuto.equals(super.languageSource)) {
				super.buttonSearch.setText(R.string.dictionarySearch);
			} else {
				super.buttonSearch.setText(String.format(super.getString(R.string.dictionaryOfflineSearchIn),
						super.languageSource));
			}
		} else {
			super.buttonSearch.setText(R.string.dictionaryOfflineStopLoading);
			super.buttonSearch.setEnabled(true);
		}
	}

	private void search(String search, boolean withLimit) {
		this.setWidgetEnabled(false);

		this.entryListFragment.clear();

		super.showProgressDialog();

		final Language lang = this.languageAuto.equals(super.languageSource) ? null : super.languageSource;
		this.stopDictionaryThread(true);
		DictionaryOfflineActivity.thread = new DictionaryThread(Thread.currentThread(), this.handler, search, lang,
				withLimit);
		DictionaryOfflineActivity.thread.start();
	}

	/**
	 * Prevent something like an SQL injection.
	 * 
	 * @param text
	 *            The text to put in the LIKE clause
	 * @return the secured text
	 */
	private static String secureTextMatch(String text) {
		int repArraySize = DictionaryOfflineActivity.TEXT_REPLACE.length;
		int txtLen = text.length();
		StringBuilder result = new StringBuilder(txtLen);

		for (int i = 0; i < txtLen; i++) {
			int j = 0;
			for (; j < repArraySize; j++) {
				String txtSearch = DictionaryOfflineActivity.TEXT_REPLACE[j][0];
				int txtSearchLen = txtSearch.length();
				// regionMatches -> startWith
				if (text.regionMatches(i, txtSearch, 0, txtSearchLen)) {
					result.append(DictionaryOfflineActivity.TEXT_REPLACE[j][1]);
					i += txtSearchLen - 1;
					break;
				}
			}
			if (j == repArraySize) {
				// no replacement found, add the original text character
				result.append(text.charAt(i));
			}
		}
		String secText = result.toString().trim();

		return secText;
	}

	/**
	 * 
	 * @param wait
	 *            true if we wait the loading thread to end. false only if this activity is
	 *            finishing.
	 */
	private void stopDictionaryThread(boolean wait) {

		if (null != DictionaryOfflineActivity.thread) {

			synchronized (Thread.currentThread()) {
				try {
					if (wait) {
						DictionaryOfflineActivity.thread.requestStop();
						// wait, the DictionaryThread will send us a notify().
						Thread.currentThread().wait();
					} else {
						// This activity is finishing

						// Prevent the handler to handle another thread
						// messages.
						this.handler.finished = true;

						// Change it priority to not slow too much the current
						// activities. It will finish anyway.
						DictionaryOfflineActivity.thread.setPriority(Thread.MIN_PRIORITY);

						// Request the thread to stop
						DictionaryOfflineActivity.thread.requestStop();
					}
				} catch (InterruptedException ie) {
				}
			}
		}
	}

	/**
	 * Glue to update the UI by the DictionaryThread.
	 */
	static class DictionaryHandler extends Handler {
		static final String MSG_ADD_ITEM = "item";
		static final String MSG_END = "end";
		static final int WHAT_START = 1;
		static final int WHAT_ADD = 2;
		static final int WHAT_END = 3;

		final DictionaryOfflineActivity activity;
		/**
		 * Boolean to discard dictionary results, when this activity is finishing.
		 */
		boolean finished;
		boolean firstMessage;

		public DictionaryHandler(DictionaryOfflineActivity activity) {
			super();
			this.activity = activity;
			this.finished = false;
			this.firstMessage = false;
		}

		public void addResult(Entry entry) {
			Message message = Message.obtain(this);
			Bundle bundle = message.getData();
			bundle.clear();

			message.what = WHAT_ADD;
			bundle.putParcelable(MSG_ADD_ITEM, entry);
			super.sendMessage(message);
		}

		public void searchStart() {
			Message message = Message.obtain(this);
			message.what = WHAT_START;
			super.sendMessage(message);
		}

		/**
		 * 
		 * @param searchDone
		 *            true if the search was complete.
		 */
		public void searchEnd(boolean searchDone) {
			Message message = Message.obtain(this);
			Bundle bundle = message.getData();
			bundle.clear();
			message.what = WHAT_END;
			bundle.putBoolean(MSG_END, searchDone);
			super.sendMessage(message);
		}

		@Override
		public void handleMessage(Message msg) {
			if (this.finished) {
				DictionaryOfflineActivity.thread = null;
				return;
			}

			if (this.firstMessage) {
				this.firstMessage = false;
				this.activity.dismissProgressDialog();
				// Allow to stop the fetching of results:
				this.activity.buttonSearch.setTag(BUTTON_STOP_LOADING);
				this.activity.updateButtonSearchText();
			}

			final Bundle data = msg.getData();
			switch (msg.what) {
			case WHAT_ADD:
				Entry entry = (Entry) msg.getData().getParcelable(MSG_ADD_ITEM);
				((EntryArrayAdapter) this.activity.entryListFragment.getListAdapter()).add(entry);
				break;
			case WHAT_START:
				this.firstMessage = true;

				break;
			case WHAT_END:
				boolean searchComplete = data.getBoolean(MSG_END);

				// Frakking toaster ?
				int numResult = DictionaryOfflineActivity.thread.numResults;
				String toastMessage = null;

				if (0 == numResult) {
					toastMessage = this.activity.getString(R.string.dictionaryNoResult);
				} else if (searchComplete) {
					String toastMessageTemplate = this.activity.getString(R.string.toastNumWords);
					toastMessage = String.format(toastMessageTemplate, DictionaryOfflineActivity.thread.searchText,
							numResult);
				} else {
					String toastMessageTemplate = this.activity.getString(R.string.toastNumWordsInterrupt);
					toastMessage = String.format(toastMessageTemplate, DictionaryOfflineActivity.thread.searchText,
							numResult);
				}
				Toast.makeText(this.activity, toastMessage, Toast.LENGTH_LONG).show();

				// Save the search result
				this.activity.addSearchResult(DictionaryOfflineActivity.thread.offlineResult);

				// We can do another search
				this.activity.setWidgetEnabled(true);
				this.activity.buttonSearch.setTag(BUTTON_DO_SEARCH);
				this.activity.updateButtonSearchText();

				DictionaryOfflineActivity.thread = null;
			}

		}
	}
}
