/*
 *  Copyright (C) 2012 Yvon TANGUY
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.dictionary;

import java.util.ArrayList;

import org.vono.narau.common.Common.KanaType;
import org.vono.narau.db.DatabaseType;
import org.vono.narau.db.DictionaryDB;
import org.vono.narau.db.DictionaryDB.DictionaryIterator;
import org.vono.narau.dictionary.DictionaryOfflineActivity.DictionaryHandler;
import org.vono.narau.models.dictionary.Entry;
import org.vono.narau.models.history.OfflineResult;
import org.vono.narau.models.kanji.KanjiData;
import org.vono.narau.preferences.Preferences;
import org.vono.narau.utils.Language;

public class DictionaryThread extends Thread {
	private static final String TAG = DictionaryThread.class.getSimpleName();
	/**
	 * Number of items (Entry) to send to the main thread before sleeping a
	 * little. To let the main thread take his breath.
	 */
	private static final int LOAD_COUNT_PAUSE = 100;
	/**
	 * How many time to sleep waiting.<br/>
	 * 
	 * @see LOAD_COUNT_PAUSE
	 */
	private static final long LOAD_COUNT_SLEEP = 1000L;

	final String searchText;
	private DictionaryHandler handler;

	/**
	 * The main thread, used to synchronized.
	 */
	private final Thread mainThread;
	/**
	 * Boolean used when the activity is finishing, or the user wants to stop
	 * the loading.
	 */
	private boolean mainThreadRequestStop;
	/**
	 * Boolean used when the activity change (device rotation...).
	 */
	private boolean mainThreadWaitActivityChange;
	private boolean withLimit;
	OfflineResult offlineResult;
	Language language;
	int numResults;

	DictionaryThread(Thread callerThread, DictionaryHandler handler, String text, Language language, boolean withLimit) {
		this.mainThread = callerThread;
		this.handler = handler;
		this.searchText = text;
		this.withLimit = withLimit;
		this.language = language;
		this.mainThreadRequestStop = false;
		this.mainThreadWaitActivityChange = false;
		this.offlineResult = null;
		this.numResults = 0;

		super.setName(DictionaryThread.TAG);
	}

	public DictionaryThread(Thread callerThread, DictionaryHandler handler, OfflineResult offR) {
		this.mainThread = callerThread;
		this.handler = handler;
		this.searchText = offR.searchTerm;
		this.offlineResult = offR;
		this.mainThreadRequestStop = false;
		this.mainThreadWaitActivityChange = false;
		this.withLimit = false;
		this.numResults = 0;

		super.setName(DictionaryThread.TAG);
	}

	@Override
	public void run() {

		this.handler.searchStart();

		if (null == this.offlineResult) {
			this.searchFromText();
		} else {
			if (this.offlineResult.numResults != this.offlineResult.dictionaryIds.size()) {
				this.withLimit = true;
				this.language = this.offlineResult.sourceLang;
				this.offlineResult = null;
				this.searchFromText();
			} else {
				this.searchFromResult();
			}
		}

		this.handler.searchEnd(!this.mainThreadRequestStop);
		this.handler = null;
	}

	private void searchFromText() {
		// Search type of text (Romanji, Kana, Kanji)
		KanaType kanaType = KanaType.ROMANJI;
		if (Language.JAPANESE.equals(this.language)) {
			this.language = null;
			kanaType = KanaType.HIRAGANA;
		}
		if (null == this.language) {
			final int textLength = this.searchText.length();
			for (int i = 0; i < textLength; i++) {
				int c = this.searchText.codePointAt(i);
				if (Character.charCount(c) == 2)
					i++;

				if (KanjiData.isKana(c)) {
					kanaType = KanaType.HIRAGANA;
				} else if (KanjiData.isKanji(c)) {
					kanaType = KanaType.KANJI;
					i = textLength;
				}
			}
		}

		// Fetch data from the good database
		final DictionaryIterator iterator = new DictionaryIterator();

		boolean haveResult = false;
		if (KanaType.ROMANJI == kanaType && null == this.language) {
			ArrayList<Language> languages = Preferences.getLangsPriorities(DatabaseType.dictonary);
			for (Language lang : languages) {
				haveResult = iterator.searchToJpn(this.searchText, lang, this.withLimit);
				if (haveResult) {
					this.language = lang;
					break;
				}
			}
		} else if (KanaType.ROMANJI == kanaType && null != this.language) {
			haveResult = iterator.searchToJpn(this.searchText, this.language, this.withLimit);
		} else if (KanaType.HIRAGANA == kanaType) {
			haveResult = iterator.searchFromJpnKana(this.searchText, this.withLimit);
			this.language = Language.JAPANESE;
		} else {
			haveResult = iterator.searchFromJpnKanji(this.searchText, this.withLimit);
			this.language = Language.JAPANESE;
		}

		// Main result show loop
		this.offlineResult = new OfflineResult(this.searchText, this.language, this.numResults);
		int count = 0;
		if (haveResult) {
			while (iterator.hasNext() && this.doWaits(++count % LOAD_COUNT_PAUSE == 0)) {
				// Send the entry to the main thread
				Entry entry = iterator.next();
				this.offlineResult.dictionaryIds.add(entry.sequence);
				this.handler.addResult(entry);
			}
			iterator.close();
		}
		this.numResults = count;

		if (this.mainThreadRequestStop) {
			synchronized (this.mainThread) {
				this.mainThread.notify();
			}
		}

		iterator.close();
	}

	private void searchFromResult() {
		final ArrayList<Integer> listSequence = this.offlineResult.dictionaryIds;
		int numEntries = listSequence.size();
		this.numResults = numEntries;
		for (int i = 0; i < numEntries; i++) {
			Entry entry = DictionaryDB.loadEntry(listSequence.get(i));

			if (this.doWaits(i % LOAD_COUNT_PAUSE == 0)) {
				this.handler.addResult(entry);
			} else {
				i = numEntries;
			}
		}
		this.offlineResult.date.setTime(System.currentTimeMillis());

	}

	/**
	 * Do thread synchro.
	 * 
	 * @return true if the main loop can continue
	 */
	private boolean doWaits(boolean mainThreadSleep) {
		final boolean requestStop;
		final boolean updateActivity;

		synchronized (this) {
			requestStop = this.mainThreadRequestStop;
			updateActivity = this.mainThreadWaitActivityChange;
		}

		try {

			if (requestStop) {
				// Do nothing
			} else if (updateActivity) {
				synchronized (this) {
					// Tigh: What do you think?
					// Adama: I think we wait.
					super.wait();
				}
			} else if (mainThreadSleep) {
				Thread.sleep(LOAD_COUNT_SLEEP);
			}

		} catch (InterruptedException ie) {
			// Don't mind
		}

		return !requestStop;
	}

	public synchronized void requestStop() {
		this.mainThreadRequestStop = true;
		super.notify();
	}

	/**
	 * Make this thread to do a pause.
	 */
	public synchronized void pause() {
		this.mainThreadWaitActivityChange = true;
	}

	public synchronized void setStopWaitingActivity(DictionaryHandler handler) {
		this.handler = handler;
		this.mainThreadWaitActivityChange = false;
		super.notify();
	}

}
