/*
 *  Copyright (C) 2012 Yvon TANGUY
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.models.history;

import java.io.IOException;
import java.io.StringReader;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.security.InvalidParameterException;
import java.util.ArrayList;
import java.util.Date;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.http.HttpStatus;
import org.vono.narau.R;
import org.vono.narau.common.Common;
import org.vono.narau.utils.JsonReader;
import org.vono.narau.utils.JsonToken;
import org.vono.narau.utils.Language;

import android.os.Parcel;
import android.os.Parcelable;

/**
 * Result on an online search.
 */
public class OnlineResult extends SearchResult {

	public final String service;
	public final String html;

	public OnlineResult(Date date, String term, int numResults, Language srcLang, Language dstLang, String service,
			String html) {
		super(date, term, numResults, srcLang, dstLang);
		this.service = service;
		this.html = html;
	}

	public OnlineResult(int serviceStringId, String term, Language srcLang, Language dstLang, StringBuilder text)
			throws IOException {
		super(new Date(), term, 0, srcLang, dstLang);
		Translation translation = null;

		switch (serviceStringId) {
		case R.string.dictionaryOnlineBing:
			translation = new MicrosoftTranslation(text);
			break;
		case R.string.dictionaryOnlineGoogle:
			translation = new GoogleTranslation(text.toString());
			break;
		case R.string.dictionaryOnlineMyMemory:
			translation = new MyMemoryTranslation(text.toString());
			break;
		default:
			throw new InvalidParameterException("serviceStringId: unknown service");
		}

		super.numResults = translation.getNumResults();
		this.service = Common.getString(serviceStringId);
		this.html = translation.toString();
	}

	protected OnlineResult(Parcel in) {
		super(in);
		this.service = in.readString();
		this.html = in.readString();
	}

	@Override
	public String getDisplayLanguage() {
		return this.service + "\n" + this.sourceLang.displayName + "/" + this.destLang.displayName;
	}

	@SuppressWarnings("cast")
	@Override
	public long getId() {
		return ((long) this.searchTerm.hashCode()) + ((long) this.sourceLang.langISO3.hashCode())
				+ ((long) this.destLang.langISO3.hashCode()) + ((long) this.service.hashCode());
	}

	@Override
	public void writeToParcel(Parcel out, int flags) {
		super.writeToParcel(out, flags);
		out.writeString(this.service);
		out.writeString(this.html);
	}

	public static final Parcelable.Creator<OnlineResult> CREATOR = new Parcelable.Creator<OnlineResult>() {
		public OnlineResult createFromParcel(Parcel in) {
			return new OnlineResult(in);
		}

		public OnlineResult[] newArray(int size) {
			return new OnlineResult[size];
		}
	};

	private static abstract class Translation {
		protected String html;

		protected abstract int getNumResults();

		@Override
		public String toString() {
			return this.html;
		}
	}

	/**
	 * Parse a Microsoft Translation translation HTML result.
	 */
	private static class MicrosoftTranslation extends Translation {
		static final int BING_START_RESULT = "onDictComplete(decodeURIComponent(\"".length();
		static final int BING_END_RESULT = "\"));".length();
		static final String PATTERN_HR = "(<hr\\s*/>)";
		static final String REPLACEMENT = "<br/><br/>";
		int numResults;

		MicrosoftTranslation(StringBuilder text) throws UnsupportedEncodingException {
			if (text.length() > (BING_START_RESULT + BING_END_RESULT)) {
				text.delete(0, BING_START_RESULT);
				text.delete(text.length() - BING_END_RESULT, text.length());
				String temp = URLDecoder.decode(text.toString(), Common.CHARSET_UTF8);

				// Is there a better way to count PATTERN_HR ?
				final Pattern pattern = Pattern.compile(PATTERN_HR);
				final Matcher matcher = pattern.matcher(temp);
				this.numResults = matcher.groupCount();

				super.html = matcher.replaceAll(REPLACEMENT);

			} else {
				this.numResults = 0;
				super.html = Common.getString(R.string.dictionaryNoResult);
			}
		}

		@Override
		protected int getNumResults() {
			return this.numResults;
		}

	}

	/**
	 * Parse a MyMemory translation JSON result.
	 */
	private static class MyMemoryTranslation extends Translation {

		private ArrayList<Match> matches;
		private String text;

		MyMemoryTranslation(String text) throws IOException {
			this.matches = null;
			super.html = null;
			this.text = text;

			this.parse();
		}

		private void parse() throws IOException {
			StringReader sr = new StringReader(this.text);
			JsonReader jr = new JsonReader(sr);
			int responseStatus = 0;
			String responseDetail = null;

			jr.beginObject();
			while (jr.hasNext()) {

				String name = jr.nextName();
				if (HttpStatus.SC_OK == responseStatus && jr.peek() == JsonToken.BEGIN_ARRAY && "matches".equals(name)) {
					this.matches = new ArrayList<Match>();
					jr.beginArray();

					while (jr.hasNext()) {
						this.matches.add(new Match(jr));
					}
					jr.endArray();

				} else if ("responseStatus".equals(name)) {
					responseStatus = jr.nextInt();
				} else if ("responseDetails".equals(name)) {
					responseDetail = jr.nextString();
				} else {
					jr.skipValue();
				}

			}
			jr.endObject();
			jr.close();

			if (HttpStatus.SC_OK != responseStatus) {
				if (responseDetail != null && !responseDetail.isEmpty()) {
					this.html = responseDetail + " (" + responseStatus + ")";
				} else {
					this.html = "Error " + responseStatus;
				}
			}
		}

		@Override
		protected int getNumResults() {
			if (null != this.matches) {
				return this.matches.size();
			}
			return 0;
		}

		@Override
		public String toString() {
			if (null == super.html) {
				if (null != this.matches) {
					StringBuilder sb = new StringBuilder();

					for (Match match : this.matches) {
						sb.append(match.toString());
					}

					super.html = sb.toString();
				} else {
					super.html = Common.getString(R.string.dictionaryNoResult);
				}
			}

			return super.html;
		}

		class Match {
			long id;
			String segment;
			String translation;
			String quality;
			String reference;
			int usageCount;
			String subject;
			String createdBy;
			String lastUpdateBy;
			String createdDate;
			String lastUpdateDate;
			double match;

			public Match(JsonReader jr) throws IOException {
				this.id = -1;
				this.usageCount = -1;
				this.match = -1f;
				this.quality = null;
				this.subject = null;
				this.segment = null;
				this.translation = null;
				this.createdBy = null;
				this.lastUpdateBy = null;
				this.createdDate = null;
				this.lastUpdateDate = null;
				this.parse(jr);
			}

			private void parse(JsonReader jr) throws IOException {
				jr.beginObject();

				while (jr.hasNext()) {

					String name = jr.nextName();

					if (JsonToken.NULL == jr.peek()) {
						jr.skipValue();
						continue;
					}

					if ("id".equals(name)) {
						this.id = jr.nextLong();
					} else if ("segment".equals(name)) {
						this.segment = jr.nextString().trim();
					} else if ("translation".equals(name)) {
						this.translation = jr.nextString().trim();
					} else if ("quality".equals(name)) {
						this.quality = jr.nextString().trim();
					} else if ("reference".equals(name)) {
						this.reference = jr.nextString().trim();
					} else if ("usage-count".equals(name)) {
						this.usageCount = jr.nextInt();
					} else if ("subject".equals(name)) {
						this.subject = jr.nextString().trim();
					} else if ("created-by".equals(name)) {
						this.createdBy = jr.nextString().trim();
					} else if ("last-updated-by".equals(name)) {
						this.lastUpdateBy = jr.nextString().trim();
					} else if ("create-date".equals(name)) {
						this.createdDate = jr.nextString().trim();
					} else if ("last-update-date".equals(name)) {
						this.lastUpdateDate = jr.nextString().trim();
					} else if ("match".equals(name)) {
						this.match = jr.nextDouble();
					} else {
						jr.skipValue();
					}
				}

				jr.endObject();

			}

			@Override
			public String toString() {
				StringBuilder sb = new StringBuilder();

				sb.append("<p>");
				sb.append(this.segment);
				sb.append(":&nbsp;<b>");
				sb.append(this.translation);
				sb.append("</b><br/>Subject:&nbsp;");
				sb.append(this.subject);
				if (null != this.quality && !this.quality.isEmpty()) {
					sb.append("<br/>Quality:&nbsp;");
					sb.append(this.quality);
				}
				if (null != this.reference && !this.reference.isEmpty()) {
					sb.append("<br/>Reference:&nbsp;");
					sb.append(this.reference);
				}
				if (null != this.createdBy && !this.createdBy.isEmpty()) {
					sb.append("<br/>Created by:&nbsp;");
					sb.append(this.createdBy);
				}
				if (null != this.createdDate && !this.createdDate.isEmpty()) {
					sb.append("<br/>Created date:&nbsp;");
					sb.append(this.createdDate);
				}
				if (null != this.lastUpdateBy && !this.lastUpdateBy.isEmpty()) {
					sb.append("<br/>Last update by:&nbsp;");
					sb.append(this.lastUpdateBy);
				}
				if (null != this.lastUpdateDate && !this.lastUpdateDate.isEmpty()) {
					sb.append("<br/>Last update date:&nbsp;");
					sb.append(this.lastUpdateDate);
				}
				if (this.usageCount > 0) {
					sb.append("<br/>Usage count:&nbsp;");
					sb.append(this.usageCount);
				}
				if (this.match >= 0f) {
					sb.append("<br/>Match:&nbsp;");
					sb.append(this.match);
				}
				sb.append("<br/><i>id:&nbsp;");
				sb.append(this.id);
				sb.append("</i><br/>");

				sb.append("</p>\n");

				return sb.toString();
			}
		}

	}

	/**
	 * Parse a Google translation JSON result.
	 */
	private static class GoogleTranslation extends Translation {
		ArrayList<Sentence> sentences;
		ArrayList<Dict> dicts;

		final String text;

		GoogleTranslation(String text) throws IOException {
			this.dicts = null;
			this.sentences = null;
			this.text = text;

			this.parse();
		}

		private void parse() throws IOException {
			StringReader sr = new StringReader(this.text);
			JsonReader jr = new JsonReader(sr);

			jr.beginObject();
			while (jr.hasNext()) {
				String name = jr.nextName();
				if ("sentences".equals(name)) {
					jr.beginArray();

					this.sentences = new ArrayList<Sentence>();
					while (jr.hasNext()) {
						this.sentences.add(new Sentence(jr));
					}

					jr.endArray();
				} else if ("dict".equals(name)) {
					jr.beginArray();

					this.dicts = new ArrayList<Dict>();
					while (jr.hasNext()) {
						this.dicts.add(new Dict(jr));
					}

					jr.endArray();

				} else {
					// ignore "src", "server_time"
					jr.skipValue();
				}
			}
			jr.endObject();
			jr.close();
		}

		@Override
		protected int getNumResults() {
			// if (null != this.dict && null != this.dict.entries &&
			// !this.dict.entries.isEmpty()) {
			// Entry entry = this.dict.entries.get(0);
			// if (null != entry.reverseTranslation)
			// return entry.reverseTranslation.size();
			// return this.dict.entries.size();
			// }
			return 1;
		}

		@Override
		public String toString() {
			if (null == super.html) {
				StringBuilder sb = new StringBuilder();
				sb.append("<p>");
				if (null != this.sentences) {
					for (Sentence sentence : this.sentences) {
						sb.append(sentence.toString());
					}
				}
				sb.append("</p><p>");

				if (null != this.dicts) {
					for (Dict dict : this.dicts) {
						sb.append(dict.toString());
					}
				}
				sb.append("</p>");

				super.html = sb.toString();
			}
			return super.html;
		}

		class Sentence {
			String trans;
			String orig;
			String tanslit;
			String srcTranslit;

			public Sentence(JsonReader jr) throws IOException {
				this.parse(jr);
			}

			private void parse(JsonReader jr) throws IOException {
				jr.beginObject();

				while (jr.hasNext()) {
					String name = jr.nextName();
					if ("trans".equals(name)) {
						this.trans = jr.nextString();
					} else if ("orig".equals(name)) {
						this.orig = jr.nextString();
					} else if ("translit".equals(name)) {
						this.tanslit = jr.nextString();
					} else if ("src_translit".equals(name)) {
						this.srcTranslit = jr.nextString();
					} else {
						// unknown
						jr.skipValue();
					}
				}

				jr.endObject();
			}

			@Override
			public String toString() {
				StringBuilder sb = new StringBuilder();

				sb.append(this.orig);
				sb.append(":<br/>");
				sb.append(this.trans);
				if (null != this.tanslit && !this.tanslit.isEmpty()) {
					sb.append("&nbsp;(");
					sb.append(this.tanslit);

					if (null != this.srcTranslit && !this.srcTranslit.isEmpty()) {
						sb.append(": ");
						sb.append(this.srcTranslit);
					}
					sb.append(")");
				}
				sb.append("<br/>");

				return sb.toString();
			}
		}

		class Dict {
			String pos;
			ArrayList<String> terms;
			ArrayList<Entry> entries;

			Dict(JsonReader jr) throws IOException {
				this.pos = null;
				this.terms = null;
				this.entries = new ArrayList<Entry>();
				this.parse(jr);
			}

			void parse(JsonReader jr) throws IOException {
				jr.beginObject();

				while (jr.hasNext()) {
					String name = jr.nextName();
					if ("pos".equals(name)) {
						this.pos = jr.nextString();
					} else if ("term".equals(name)) {
						this.terms = new ArrayList<String>();
						jr.beginArray();

						while (jr.hasNext()) {
							this.terms.add(jr.nextString());
						}

						jr.endArray();
					} else if ("entry".equals(name)) {
						this.entries = new ArrayList<Entry>();
						jr.beginArray();
						while (jr.hasNext()) {
							this.entries.add(new Entry(jr));
						}
						jr.endArray();
					} else {
						// unknown
						jr.skipValue();
					}
				}

				jr.endObject();
			}

			@Override
			public String toString() {
				StringBuilder sb = new StringBuilder();

				sb.append(this.pos);
				if (null != this.terms) {
					sb.append(": ");
					for (String term : this.terms) {
						sb.append(term);
						sb.append("&nbsp");
					}
				}
				sb.append("<br/>");
				if (null != this.entries) {
					for (Entry entry : this.entries) {
						sb.append(entry.toString());
					}
				}

				sb.append("<br/>");
				return sb.toString();
			}

		}

		class Entry {
			String word;
			ArrayList<String> reverseTranslation;
			double score;

			Entry(JsonReader jr) throws IOException {
				this.word = null;
				this.score = 0f;
				this.reverseTranslation = new ArrayList<String>();
				this.parse(jr);
			}

			void parse(JsonReader jr) throws IOException {
				jr.beginObject();

				while (jr.hasNext()) {
					String name = jr.nextName();

					if ("word".equals(name)) {
						this.word = jr.nextString();
					} else if ("reverse_translation".equals(name)) {
						this.reverseTranslation = new ArrayList<String>();
						jr.beginArray();
						while (jr.hasNext()) {
							this.reverseTranslation.add(jr.nextString());
						}
						jr.endArray();
					} else if ("score".equals(name)) {
						this.score = jr.nextDouble();
					} else {
						// "synset_id"
						jr.skipValue();
					}

				}

				jr.endObject();
			}

			@Override
			public String toString() {
				StringBuilder sb = new StringBuilder();

				sb.append(this.word);
				sb.append("&nbsp;(");
				sb.append(String.format("%1$2d%%", (int) (this.score * 100f)));
				sb.append("):&nbsp;");
				if (null != this.reverseTranslation) {
					for (String rt : this.reverseTranslation) {
						sb.append(rt);
						sb.append("&nbsp;");
					}
				}
				sb.append("<br/>");
				return sb.toString();
			}

		}
	}

}
