/*
 *  Copyright (C) 2012 Yvon TANGUY
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package android.compat.view;

import android.app.Activity;
import android.compat.view.menu.IconMenuView;
import android.compat.view.menu.MenuBuilder;
import android.compat.view.menu.SubMenuBuilder;
import android.content.Context;
import android.graphics.PixelFormat;
import android.util.SparseArray;
import android.view.Gravity;
import android.view.KeyEvent;
import android.view.Menu;
import android.view.MenuItem;
import android.view.MotionEvent;
import android.view.ViewGroup;
import android.view.WindowManager;
import android.widget.FrameLayout;

/**
 * Manage menu action, to keep the activities as close as possible from the original code.<br/>
 * Theses actions are normally perform by the Android classes (PhoneWindow...), but since they are
 * not accessible to us, poor developers, I had to implement this myself.<br/>
 * There is some trouble in the menu layout for Android 1.5 & 1.6. But is there anyone who will use
 * this app on theses phones?
 */
public class MenuCompat {
	// private static final String TAG = MenuCompat.class.getSimpleName();

	private static MenuCompat instance = null;

	private SparseArray<MenuInfos> saMenuInfos;

	private static MenuCompat getInstance() {
		if (null == MenuCompat.instance) {
			MenuCompat.instance = new MenuCompat();
		}

		return MenuCompat.instance;
	}

	private MenuCompat() {
		this.saMenuInfos = new SparseArray<MenuCompat.MenuInfos>();
	}

	/**
	 * 
	 * @param activity
	 *            caller activity
	 * @param keyCode
	 *            current keycode
	 * @param menuId
	 *            the menu id
	 * @return the menu if display, or null if hide
	 */
	public static Menu onKeyUp(Activity activity, int keyCode, int menuId) {
		return MenuCompat.getInstance().doKeyUp(activity, keyCode, menuId);
	}

	/**
	 * Prevent window leak when configuration change. Need to be called during the onStop() event in
	 * the activity.
	 * 
	 * @param menuId
	 *            the menu resource id.
	 */
	public static void closeMenu(int menuId) {
		MenuInfos mi = MenuCompat.getInstance().saMenuInfos.get(menuId);
		if (null != mi && mi.visible) {
			mi.menuBuilder.close();
		}
	}

	private Menu doKeyUp(Activity activity, int keyCode, int menuId) {
		MenuInfos mi = this.saMenuInfos.get(menuId);
		if (null == mi) {
			mi = new MenuInfos();
			this.saMenuInfos.put(menuId, mi);
		}

		if (keyCode == KeyEvent.KEYCODE_MENU) {
			return mi.showHideMenu(activity, menuId);
		}

		return null;
	}

	@SuppressWarnings("unqualified-field-access")
	private static class MenuInfos implements MenuBuilder.Callback {
		private MenuBuilder menuBuilder;
		private DecorView frameLayout;
		private WindowManager.LayoutParams wmLP;
		private IconMenuView menuView;

		private Activity mActivity;

		boolean visible;

		MenuInfos() {
			this.menuBuilder = null;
			this.visible = false;
			this.frameLayout = null;
			this.wmLP = null;
			this.mActivity = null;
		}

		Menu showHideMenu(Activity activity, int menuId) {

			if (activity != this.mActivity) {
				// the frame layout may be attach to the previous activity
				// so we have to rebuild it, to avoid an exception during
				// WindowManager.addView() call
				// The is still a window leak when the configuration change (device rotation), but
				// this do not lead to a crash. I do not figure how to prevent this without closing
				// the menu in each activities, during the call of onStop().

				if (null != this.frameLayout) {
					this.frameLayout.removeView(this.menuView);
					if (null != this.frameLayout.getParent()) {
						((ViewGroup) this.frameLayout.getParent()).removeView(this.frameLayout);
					}
				}

				this.menuBuilder = null;
				this.frameLayout = null;
				this.visible = false;
			}

			if (this.visible) {
				this.menuBuilder.close();
				return null;
			}

			this.mActivity = activity;

			final WindowManager wm = activity.getWindowManager();
			if (null == wm) {
				return null;
			}

			this.visible = true;

			if (null == this.menuBuilder) {
				this.menuBuilder = new MenuBuilder(activity.getApplicationContext());
				MenuInflater inflater = new MenuInflater(activity.getApplicationContext());
				inflater.inflate(menuId, this.menuBuilder);

				this.menuBuilder.setCallback(this);
			}

			if (null == this.frameLayout) {

				this.frameLayout = new DecorView(activity.getApplicationContext(), this);
				this.menuView = (IconMenuView) this.menuBuilder.getMenuView(MenuBuilder.TYPE_ICON, this.frameLayout);

				FrameLayout.LayoutParams vglp = new FrameLayout.LayoutParams(ViewGroup.LayoutParams.WRAP_CONTENT,
						ViewGroup.LayoutParams.WRAP_CONTENT);
				// vglp.gravity = Gravity.CENTER | Gravity.BOTTOM;

				if (this.menuView.getParent() != null) {
					((ViewGroup) this.menuView.getParent()).removeView(this.menuView);
				}
				this.frameLayout.addView(this.menuView, vglp);

				int x = 0, y = 0;
				this.wmLP = new WindowManager.LayoutParams(ViewGroup.LayoutParams.WRAP_CONTENT,
						ViewGroup.LayoutParams.WRAP_CONTENT, x, y,
						WindowManager.LayoutParams.TYPE_APPLICATION_ATTACHED_DIALOG,
						WindowManager.LayoutParams.FLAG_DITHER | WindowManager.LayoutParams.FLAG_ALT_FOCUSABLE_IM,
						PixelFormat.TRANSLUCENT);

				this.wmLP.gravity = Gravity.CENTER | Gravity.BOTTOM;
				this.wmLP.windowAnimations = this.menuView.getWindowAnimations();
			}

			wm.addView(this.frameLayout, this.wmLP);

			this.frameLayout.requestFocus();
			this.menuView.requestLayout();

			return this.menuBuilder;
		}

		public boolean onMenuItemSelected(MenuBuilder menu, MenuItem item) {
			// Log.d(TAG, "onMenuItemSelected");
			return mActivity.onOptionsItemSelected(item);
		}

		public void onCloseMenu(MenuBuilder menu, boolean allMenusAreClosing) {
			// Log.d(TAG, "onCloseMenu");

			if (visible) {
				this.visible = false;
				final WindowManager wm = this.mActivity.getWindowManager();
				wm.removeView(this.frameLayout);
			}
		}

		public boolean onSubMenuSelected(SubMenuBuilder subMenu) {
			// Log.d(TAG, "onSubMenuSelected");
			return false;
		}

		public void onCloseSubMenu(SubMenuBuilder menu) {
			// Log.d(TAG, "onCloseSubMenu");

		}

		public void onMenuModeChange(MenuBuilder menu) {
			// Log.d(TAG, "onMenuModeChange");
		}

	}

	private static class DecorView extends FrameLayout {
		private MenuInfos menuInfos;

		public DecorView(Context context, MenuInfos infos) {
			super(context);
			this.menuInfos = infos;
		}

		@Override
		public boolean onTouchEvent(MotionEvent event) {
			if (!super.onTouchEvent(event)) {
				// close the menu if the user click outside the view.
				this.menuInfos.menuBuilder.close();
			}
			return true;
		}

		@Override
		public boolean onKeyUp(int keyCode, KeyEvent event) {
			// Log.d(TAG, "MenuWindow.onKeyUp" + keyCode);
			if (keyCode == KeyEvent.KEYCODE_BACK || keyCode == KeyEvent.KEYCODE_MENU) {
				this.menuInfos.menuBuilder.close();
				return true;
			}
			return super.onKeyUp(keyCode, event);
		}

		@Override
		public boolean dispatchKeyEvent(KeyEvent event) {
			// Log.d(TAG, "MenuWindow.dispatchKeyEvent" + event.getAction());

			switch (event.getAction()) {
			case KeyEvent.ACTION_UP:
				return this.onKeyUp(event.getKeyCode(), event);
			}

			return super.dispatchKeyEvent(event);
		}

	}
}
