/*
 *  Copyright (C) 2012 Yvon TANGUY
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.dictionary;

import org.vono.narau.R;
import org.vono.narau.common.Common;
import org.vono.narau.common.Common.KanaType;
import org.vono.narau.db.HistoryDB;
import org.vono.narau.dictionary.HistoryFragment.OnSearchSelectedListener;
import org.vono.narau.models.history.OfflineResult;
import org.vono.narau.models.history.OnlineResult;
import org.vono.narau.models.history.SearchResult;
import org.vono.narau.preferences.Preferences;
import org.vono.narau.utils.Language;

import android.app.ProgressDialog;
import android.content.Context;
import android.content.Intent;
import android.content.res.Configuration;
import android.os.Bundle;
import android.support.v4.app.FragmentActivity;
import android.view.MenuItem;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.View.OnLongClickListener;
import android.view.inputmethod.InputMethodManager;
import android.widget.Button;
import android.widget.EditText;
import android.widget.ImageButton;
import android.widget.LinearLayout;

/**
 * Base class for offline & online dictionary search.
 * 
 */
abstract class DictionaryBase extends FragmentActivity implements OnLongClickListener, OnSearchSelectedListener,
		OnClickListener {

	protected static final String KEY_TXT_SEARCH = "DB_TXT";
	protected static final String KEY_BUTTON_CONV = "DB_B_CONV";
	protected static final String KEY_LANGUAGE_SRC = "DB_LANG_SOURCE";
	protected static final String KEY_LANGUAGE_DST = "DB_LANG_DEST";

	private static final char A_ROMANJI = 'a';
	private static final char A_HIRAGANA = 'あ';
	private static final char A_KATAKANA = 'ア';

	protected static final String LANG_PAIR_SEP = ",";

	private static final int ACTIVITY_HISTORY_RESULT_CODE = 1209;

	protected Button buttonConvRomanji;
	protected Button buttonSearch;
	protected ImageButton buttonClear;
	protected EditText editTextInput;
	private LinearLayout linearLayoutZoomControls;

	private ProgressDialog progressDialog;

	protected final DictionaryType type;

	protected Language languageSource;
	protected Language languageDest;

	public DictionaryBase(DictionaryType type) {
		super();

		this.buttonConvRomanji = null;
		this.buttonSearch = null;
		this.editTextInput = null;
		this.progressDialog = null;
		this.type = type;
		this.languageSource = null;
		this.languageDest = null;
	}

	@Override
	public void setContentView(int layoutResID) {
		super.setContentView(layoutResID);

		this.editTextInput = (EditText) super.findViewById(R.id.dictionaryEditTextInput);
		this.buttonConvRomanji = (Button) super.findViewById(R.id.dictionaryButtonConvRomanji);
		this.buttonSearch = (Button) super.findViewById(R.id.dictionaryButtonSearch);
		this.buttonClear = (ImageButton) super.findViewById(R.id.dictionaryButtonReset);
		this.linearLayoutZoomControls = (LinearLayout) super.findViewById(R.id.zoomControlsLinearLayout);

		this.buttonClear.setOnClickListener(this);
		this.buttonConvRomanji.setOnClickListener(this);
		this.buttonSearch.setOnClickListener(this);

		((Button) super.findViewById(R.id.zoomControlsButtonIn)).setOnClickListener(this);
		((Button) super.findViewById(R.id.zoomControlsButtonOut)).setOnClickListener(this);
	}

	@Override
	protected void onSaveInstanceState(Bundle outState) {
		super.onSaveInstanceState(outState);

		outState.putChar(KEY_BUTTON_CONV, this.buttonConvRomanji.getText().charAt(0));
		outState.putCharSequence(KEY_TXT_SEARCH, this.editTextInput.getText());
		outState.putParcelable(KEY_LANGUAGE_SRC, this.languageSource);
		outState.putParcelable(KEY_LANGUAGE_DST, this.languageDest);
	}

	@Override
	protected void onPostCreate(Bundle bundle) {
		super.onPostCreate(bundle);

		if (bundle.containsKey(KEY_BUTTON_CONV)) {
			this.buttonConvRomanji.setText(Character.valueOf(bundle.getChar(KEY_BUTTON_CONV)).toString());
		} else {
			this.buttonConvRomanji.setText(Character.valueOf(A_HIRAGANA).toString());
		}

		if (bundle.containsKey(KEY_TXT_SEARCH)) {
			this.editTextInput.setText(bundle.getCharSequence(KEY_TXT_SEARCH));
		}

		final String langPair;
		switch (this.type) {
		case offline:
			langPair = Preferences.getString(R.string.prefNameDictionaryOffLineLangPair);
			break;
		case online:
		default:
			langPair = Preferences.getString(R.string.prefNameDictionaryOnLineLangPair);
			break;
		}

		final String[] langs;
		if (null != langPair) {
			langs = langPair.split(LANG_PAIR_SEP);
		} else {
			langs = null;
		}

		if (bundle.containsKey(KEY_LANGUAGE_SRC)) {
			this.languageSource = bundle.getParcelable(KEY_LANGUAGE_SRC);
		} else if (null != langs && langs.length == 2) {
			this.languageSource = Language.MAP_ISO3.get(langs[0]);
		} // else let the super class choose

		if (bundle.containsKey(KEY_LANGUAGE_DST)) {
			this.languageDest = bundle.getParcelable(KEY_LANGUAGE_DST);
		} else if (null != langs && langs.length == 2) {
			this.languageDest = Language.MAP_ISO3.get(langs[1]);
		} // else let the super class choose

		this.editTextInput.requestFocus();
	}

	@Override
	protected void onDestroy() {
		super.onDestroy();

		if (super.isFinishing()) {
			this.buttonConvRomanji = null;
			this.buttonSearch = null;
			this.editTextInput = null;
			this.linearLayoutZoomControls = null;
		}
	}

	/**
	 * Click on the search button
	 * 
	 * @param button
	 *            The search button
	 */
	public abstract void onClickSearch(View button);

	@SuppressWarnings("static-method")
	public void onClickZoom(final View buttonZoom) {
		// float value = Float.NaN;
		final int id = buttonZoom.getId();
		if (R.id.zoomControlsButtonOut == id) {
			Preferences.zoomOut();
		} else if (R.id.zoomControlsButtonIn == id) {
			Preferences.zoomIn();
		} // else { error but ignore, should not append }
	}

	/**
	 * Show the result of an previous search (when the screen is too small to display the older
	 * searches), return by HistoryActivity.
	 * <p/>
	 * {@inheritDoc}
	 * 
	 * @see HistoryActivity
	 */
	@Override
	protected void onActivityResult(int requestCode, int resultCode, Intent data) {
		super.onActivityResult(requestCode, resultCode, data);

		if (ACTIVITY_HISTORY_RESULT_CODE == requestCode && RESULT_OK == resultCode) {
			SearchResult sr = data.getParcelableExtra(HistoryActivity.KEY_RESULT_SEARCH);
			this.onHistorySearchSelected(sr);
		}
	}

	@Override
	public boolean onOptionsItemSelected(MenuItem item) {
		switch (item.getItemId()) {
		case R.id.menuDictionaryShowControls:
			if (View.VISIBLE == this.linearLayoutZoomControls.getVisibility()) {
				this.linearLayoutZoomControls.setVisibility(View.GONE);
				item.setTitle(R.string.menuShowZoom);
			} else {
				this.linearLayoutZoomControls.setVisibility(View.VISIBLE);
				item.setTitle(R.string.menuHideZoom);
			}
			break;
		case R.id.menuDictionaryHistory:
			this.showHistory();
			break;
		default:
			return super.onOptionsItemSelected(item);
		}

		return true;
	}

	/**
	 * Characters conversion: romanji->hiragana->katakana
	 * 
	 * @param button
	 */
	public void onClickConvCharacters(View button) {
		char currentTextFormat = this.buttonConvRomanji.getText().charAt(0);

		final String searchText = this.editTextInput.getText().toString();
		String newSearchText = searchText;

		switch (currentTextFormat) {
		case A_HIRAGANA:
			// romanji -> hiragana
			newSearchText = Common.convertKana(searchText, KanaType.ROMANJI, KanaType.HIRAGANA);
			this.buttonConvRomanji.setText(Character.valueOf(A_KATAKANA).toString());
			break;
		case A_KATAKANA:
			// hiragana -> katakana
			newSearchText = Common.convertKana(searchText, KanaType.HIRAGANA, KanaType.KATAKANA);
			this.buttonConvRomanji.setText(Character.valueOf(A_ROMANJI).toString());
			break;
		case A_ROMANJI:
			// katakana -> romanji
			newSearchText = Common.convertKana(searchText, KanaType.KATAKANA, KanaType.ROMANJI);
			this.buttonConvRomanji.setText(Character.valueOf(A_HIRAGANA).toString());
			break;
		}

		this.editTextInput.setText(newSearchText);
	}

	/**
	 * Click on the reset button
	 * 
	 * @param button
	 *            The reset button
	 */
	public void onClickReset(View button) {

		this.editTextInput.setText(Common.EMPTY_STRING);
		this.buttonConvRomanji.setText(Character.valueOf(A_HIRAGANA).toString());
		this.editTextInput.requestFocus();
		this.setWidgetEnabled(true);

		this.switchKeyboardVisible();
	}

	protected void switchKeyboardVisible() {
		if (super.getResources().getConfiguration().keyboardHidden == Configuration.KEYBOARDHIDDEN_NO) {
			// Hide the keyboard
			InputMethodManager imm = (InputMethodManager) getSystemService(Context.INPUT_METHOD_SERVICE);
			imm.hideSoftInputFromWindow(this.editTextInput.getWindowToken(), 0);
		} else {
			InputMethodManager imm = (InputMethodManager) getSystemService(Context.INPUT_METHOD_SERVICE);
			imm.showSoftInput(this.editTextInput, 0);
		}
	}

	public DictionaryType getDictionaryType() {
		return this.type;
	}

	private void showHistory() {
		Intent intent = new Intent(this, HistoryActivity.class);
		intent.putExtra(HistoryFragment.KEY_DICTIONARY_TYPE, this.type);
		super.startActivityForResult(intent, ACTIVITY_HISTORY_RESULT_CODE);

	}

	protected void showProgressDialog() {
		if (null == this.progressDialog) {
			this.progressDialog = ProgressDialog.show(this, Common.EMPTY_STRING, super.getString(R.string.loading),
					true);
		}
	}

	protected void dismissProgressDialog() {
		if (null != this.progressDialog) {
			this.progressDialog.dismiss();
			this.progressDialog = null;
		}
	}

	protected void setWidgetEnabled(boolean value) {
		this.buttonSearch.setEnabled(value);
		this.buttonConvRomanji.setEnabled(value);
		this.editTextInput.setEnabled(value);
		this.buttonClear.setEnabled(value);

		if (value) {
			this.editTextInput.requestFocus();
		}
	}

	@SuppressWarnings("static-method")
	protected void addSearchResult(SearchResult sr) {
		if (sr instanceof OfflineResult) {
			HistoryDB.getHistoryDB().addOffileResult((OfflineResult) sr);
		} else if (sr instanceof OnlineResult) {
			HistoryDB.getHistoryDB().addOnlineResult((OnlineResult) sr);
		}// else bug ^^
	}

	protected abstract void updateButtonSearchText();

	public void onClick(View v) {
		switch (v.getId()) {
		case R.id.dictionaryButtonConvRomanji:
			this.onClickConvCharacters(v);
			break;
		case R.id.dictionaryButtonSearch:
			this.onClickSearch(v);
			break;
		case R.id.dictionaryButtonReset:
			this.onClickReset(v);
			break;
		case R.id.zoomControlsButtonIn:
		case R.id.zoomControlsButtonOut:
			this.onClickZoom(v);
			break;
		default:
		}
	}

}
