/*
 *  Copyright (C) 2012 Yvon TANGUY
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.dictionary;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.CookieStore;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.cookie.Cookie;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.params.CoreProtocolPNames;
import org.vono.narau.R;
import org.vono.narau.common.Common;
import org.vono.narau.db.HistoryDB;
import org.vono.narau.models.history.OnlineResult;
import org.vono.narau.models.history.SearchResult;
import org.vono.narau.preferences.Preferences;
import org.vono.narau.utils.Language;

import android.app.AlertDialog;
import android.compat.view.MenuCompat;
import android.content.DialogInterface;
import android.content.DialogInterface.OnClickListener;
import android.os.AsyncTask;
import android.os.Bundle;
import android.text.Html;
import android.util.Log;
import android.util.SparseArray;
import android.view.KeyEvent;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.widget.Button;
import android.widget.ImageButton;
import android.widget.TextView;

/**
 * Do a translation via:
 * <ul>
 * <li><a href="http://mymemory.translated.net/">MyMemory</a></li>
 * <li><a href="http://www.bing.com/translator/">Microsoft Translator ®</a></li>
 * <li><a href="http://translate.google.com/m/translate">Google Translation</a></li>
 * </ul>
 * 
 */
public class DictionaryOnlineActivity extends DictionaryBase implements OnClickListener {
	private static final String TAG = DictionaryOnlineActivity.class.getSimpleName();

	private static final int TAG_CHOOSE_DICT = R.string.dictionaryOnlineSearchWith;
	private static final int TAG_CHOOSE_LANG_SRC = R.string.dictionaryOnlineLangSource;
	private static final int TAG_CHOOSE_LANG_DST = R.string.dictionaryOnlineLangTarget;

	// "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:17.0) Gecko/20100101 Firefox/17.0";
	private static final String USER_AGENT = System.getProperty("http.agent");

	private static final String KEY_SERVICE_ID = "KEY_SERVICE_ID";
	private static final String KEY_HTML_RESULT = "KEY_HTML_RESULT";

	/**
	 * Services available.
	 */
	static final int[] SERVICES = { R.string.dictionaryOnlineMyMemory, R.string.dictionaryOnlineBing,
			R.string.dictionaryOnlineGoogle };

	/**
	 * Online service queries<br/>
	 * Queries format:
	 * <ul>
	 * <li>0: Cookie fetch URL</li>
	 * <li>1: scheme</li>
	 * <li>2: authority</li>
	 * <li>3: path</li>
	 * <li>4: query</li>
	 * <li>5: fragment</li>
	 * </ul>
	 * 
	 * @see URI
	 */
	private static final SparseArray<String[]> SERVICE_QUERIES;

	static {
		SERVICE_QUERIES = new SparseArray<String[]>(SERVICES.length);
		SERVICE_QUERIES.put(R.string.dictionaryOnlineMyMemory, new String[] { null, "http", "mymemory.translated.net",
				"/api/get", "q=%1$s&langpair=%2$s|%3$s", null });
		SERVICE_QUERIES.put(R.string.dictionaryOnlineBing, new String[] { null, "http", "www.microsofttranslator.com",
				"/dictionary.ashx", "from=%2$s&to=%3$s&text=%1$s&oncomplete=onDictComplete", null });
		SERVICE_QUERIES.put(R.string.dictionaryOnlineGoogle, new String[] { "http://translate.google.com/m/translate",
				"http", "translate.google.com", "/translate_a/t", "client=webapp&sl=%2$s&tl=%3$s&q=%1$s&sc=1", null });
	}

	private int serviceId;

	private Button buttonLangSource;
	private Button buttonLangDest;
	private ImageButton buttonLangSwap;
	private TextView textViewResult;

	private HashMap<String, Integer> servicesNames;

	public DictionaryOnlineActivity() {
		super(DictionaryType.online);
	}

	public static boolean haveDatabase() {
		// Method used by Narau.showMissingDatabaseNotification()
		return true;
	}

	@Override
	public void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);

		this.servicesNames = new HashMap<String, Integer>(SERVICES.length);
		for (int service : SERVICES) {
			this.servicesNames.put(super.getString(service), service);
		}

		super.setContentView(R.layout.dictionary_online);
		this.textViewResult = (TextView) super.findViewById(R.id.dictionaryOnlineTextResult);
		this.buttonLangSource = (Button) super.findViewById(R.id.dictionaryOnlineButtonLanguageSource);
		this.buttonLangDest = (Button) super.findViewById(R.id.dictionaryOnlineButtonLanguageDest);
		this.buttonLangSwap = (ImageButton) super.findViewById(R.id.dictionaryOnlineButtonLanguageSwap);

		this.buttonLangSource.setOnClickListener(this);
		this.buttonLangDest.setOnClickListener(this);
		this.buttonLangSwap.setOnClickListener(this);
		super.buttonSearch.setOnLongClickListener(this);
	}

	@Override
	protected void onSaveInstanceState(Bundle outState) {
		super.onSaveInstanceState(outState);

		outState.putInt(KEY_SERVICE_ID, this.serviceId);
		outState.putCharSequence(KEY_HTML_RESULT, this.textViewResult.getText());
	}

	@Override
	protected void onPostCreate(Bundle savedInstanceState) {

		Bundle bundle = savedInstanceState;
		if (null == bundle) {
			bundle = super.getIntent().getExtras();

			if (null == bundle) {
				// Empty bundle
				bundle = new Bundle();
			}
		}

		super.onPostCreate(bundle);

		if (bundle.containsKey(KEY_SERVICE_ID)) {
			this.serviceId = bundle.getInt(KEY_SERVICE_ID);
		} else {
			int service = Preferences.getInt(R.string.prefNameDictionaryOnLineService);
			if (0 != service && SERVICE_QUERIES.indexOfKey(service) >= 0) {
				this.serviceId = service;
			} else {
				this.serviceId = SERVICES[0];
			}
		}

		this.textViewResult.setText(bundle.getCharSequence(KEY_HTML_RESULT));

		// Default values
		if (null == this.languageSource) {
			this.languageSource = Language.DEFAULT;
		}
		if (null == this.languageDest) {
			this.languageDest = Language.JAPANESE;
		}

		this.updateCurrentLangPair();
		this.updateButtonSearchText();

		this.textViewResult.setTextSize(Preferences.getZoom());
	}

	/**
	 * {@inheritDoc} Show a dialog to choose the online service to use.
	 * 
	 * @param v
	 *            {@inheritDoc} May be null (not used).
	 */
	public boolean onLongClick(View v) {
		AlertDialog.Builder builder = new AlertDialog.Builder(this);

		String[] parameters = new String[SERVICES.length];
		int checkedItem = 0;
		for (int i = 0; i < parameters.length; i++) {
			parameters[i] = super.getString(SERVICES[i]);
			if (SERVICES[i] == this.serviceId) {
				checkedItem = i;
			}
		}
		builder.setSingleChoiceItems(parameters, checkedItem, this);

		builder.show().getListView().setTag(TAG_CHOOSE_DICT);
		return true;
	}

	/**
	 * Call after select an item in one of the dialog list. {@inheritDoc}
	 */
	public void onClick(DialogInterface dialog, int which) {
		dialog.dismiss();

		int tag = (Integer) ((AlertDialog) dialog).getListView().getTag();

		switch (tag) {
		case TAG_CHOOSE_DICT:
			this.serviceId = SERVICES[which];
			Preferences.setInt(R.string.prefNameDictionaryOnLineService, this.serviceId);
			this.updateButtonSearchText();
			break;
		case TAG_CHOOSE_LANG_SRC:
			super.languageSource = Language.LIST.get(which);
			this.updateCurrentLangPair();
			break;
		case TAG_CHOOSE_LANG_DST:
			super.languageDest = Language.LIST.get(which);
			this.updateCurrentLangPair();
			break;
		}

	}

	@Override
	public void onClickSearch(View button) {

		if (this.editTextInput.getText().length() > 0 && this.serviceId != R.string.dictionaryOnlineUnknown) {
			this.setWidgetEnabled(false);
			this.textViewResult.setText(Common.EMPTY_STRING);

			super.switchKeyboardVisible();

			GetTranslationTask task = new GetTranslationTask();
			task.execute(this.serviceId);
		}
	}

	/**
	 * 
	 * @param button
	 *            click on that button
	 */
	public void onClickButtonLanguageSource(View button) {
		this.showDialogChooseLang(TAG_CHOOSE_LANG_SRC);
	}

	/**
	 * 
	 * @param button
	 *            click on that button
	 */
	public void onClickSwapLanguages(View button) {
		Language tmp = super.languageDest;
		super.languageDest = super.languageSource;
		super.languageSource = tmp;
		this.updateCurrentLangPair();
	}

	/**
	 * 
	 * @param button
	 *            click on that button
	 */
	public void onClickButtonLanguageDest(View button) {
		this.showDialogChooseLang(TAG_CHOOSE_LANG_DST);
	}

	public void onHistorySearchSelected(SearchResult sr) {
		final OnlineResult onlineResult = (OnlineResult) sr;

		final Integer service = this.servicesNames.get(onlineResult.service);
		String buttonSearchText = null;
		if (null == service) {
			this.serviceId = R.string.dictionaryOnlineUnknown;
		} else {
			this.serviceId = service;
		}

		super.buttonSearch.setText(buttonSearchText);
		super.languageSource = onlineResult.sourceLang;
		super.languageDest = onlineResult.destLang;
		this.textViewResult.setText(Html.fromHtml(onlineResult.html));
		this.editTextInput.setText(onlineResult.searchTerm);
		this.updateCurrentLangPair();
		this.updateButtonSearchText();
	}

	@Override
	protected void onStop() {
		MenuCompat.closeMenu(R.menu.dictionary_online);
		super.onStop();
	}

	@Override
	public boolean onKeyUp(int keyCode, KeyEvent event) {
		Menu menu = MenuCompat.onKeyUp(this, keyCode, R.menu.dictionary_online);
		if (null != menu) {
			super.switchKeyboardVisible();
			return true;
		}
		return super.onKeyUp(keyCode, event);
	}

	// @Override
	// public boolean onCreateOptionsMenu(Menu menu) {
	// MenuInflater inflater = getMenuInflater();
	// inflater.inflate(R.menu.dictionary_online, menu);
	// return true;
	// }

	@Override
	public boolean onOptionsItemSelected(MenuItem item) {
		switch (item.getItemId()) {
		case R.id.menuDictionarySelectService:
			this.onLongClick(null);
			break;
		default:
			return super.onOptionsItemSelected(item);
		}

		return true;
	}

	/**
	 * Change the text size.
	 * 
	 * @param buttonZoom
	 */
	@Override
	public void onClickZoom(final View buttonZoom) {
		super.onClickZoom(buttonZoom);
		this.textViewResult.setTextSize(Preferences.getZoom());
	}

	/**
	 * Show a dialog with all available languages.
	 * 
	 * @param title
	 *            the online service string id.
	 */
	private void showDialogChooseLang(int title) {
		AlertDialog.Builder builder = new AlertDialog.Builder(this);

		final ArrayList<Language> languages = Language.LIST;

		int checkedItem = 0;
		if (title == TAG_CHOOSE_LANG_SRC) {
			checkedItem = languages.indexOf(super.languageSource);
		} else if (title == TAG_CHOOSE_LANG_DST) {
			checkedItem = languages.indexOf(super.languageDest);
		}

		builder.setTitle(title);
		builder.setSingleChoiceItems(languages.toArray(new Language[0]), checkedItem, this);

		builder.show().getListView().setTag(title);
	}

	@Override
	protected void setWidgetEnabled(boolean value) {
		super.setWidgetEnabled(value);

		if (null != this.buttonLangSwap) {
			this.buttonLangDest.setEnabled(value);
			this.buttonLangSource.setEnabled(value);
			this.buttonLangSwap.setEnabled(value);
		}
	}

	/**
	 * Display the current language pair.
	 */
	private void updateCurrentLangPair() {
		this.buttonLangSource.setText(super.languageSource);
		this.buttonLangDest.setText(super.languageDest);

		Preferences.setString(R.string.prefNameDictionaryOnLineLangPair, super.languageSource.langISO3 + LANG_PAIR_SEP
				+ super.languageDest.langISO3);
	}

	/**
	 * Change the current online service.
	 * 
	 * @see ONLINE_DICTS
	 */
	@Override
	protected void updateButtonSearchText() {
		if (R.string.dictionaryOnlineUnknown == this.serviceId) {
			super.buttonSearch.setText(super.getString(R.string.dictionaryOnlineUnknown));
		} else {
			final String format = super.getString(R.string.dictionaryOnlineSearchWith);
			super.buttonSearch.setText(String.format(format, super.getString(this.serviceId)));
		}
	}

	private static StringBuilder getContent(InputStream iStream) throws IOException {
		StringBuilder sb = new StringBuilder();

		final int bufSize = 512;
		BufferedReader br = new BufferedReader(new InputStreamReader(iStream, "UTF-8"), bufSize);
		final char[] buffer = new char[bufSize];

		int read = br.read(buffer);

		while (read != -1) {
			sb.append(buffer, 0, read);
			read = br.read(buffer);
		}

		br.close();

		return sb;
	}

	private class GetTranslationTask extends AsyncTask<Integer, Integer, StringBuilder> {

		private int serviceStringId;

		@Override
		protected StringBuilder doInBackground(Integer... params) {
			if (params.length != 1) {
				return null;
			}
			this.serviceStringId = params[0];

			StringBuilder result = null;

			try {
				result = this.getRawTranslation();
			} catch (ClientProtocolException cpe) {
				Log.e(TAG, cpe.getMessage(), cpe);
				result = new StringBuilder(cpe.getMessage());
			} catch (IOException ioe) {
				Log.e(TAG, ioe.getMessage(), ioe);
				result = new StringBuilder(ioe.getMessage());
			} catch (URISyntaxException use) {
				Log.e(TAG, use.getMessage(), use);
				result = new StringBuilder(use.getMessage());
			}

			return result;
		}

		/**
		 * Fetch the translation from the service.<br/>
		 * Return an error message if something failed. This message can be an HTML text send by the
		 * server.
		 * 
		 * @return the translation content, as return by the server.
		 * @throws ClientProtocolException
		 * @throws IOException
		 * @throws URISyntaxException
		 */
		private StringBuilder getRawTranslation() throws ClientProtocolException, IOException, URISyntaxException {

			DefaultHttpClient httpClient = new DefaultHttpClient();
			httpClient.getParams().setParameter(CoreProtocolPNames.USER_AGENT, DictionaryOnlineActivity.USER_AGENT);
			httpClient.getParams().setParameter(CoreProtocolPNames.HTTP_ELEMENT_CHARSET, Common.CHARSET_UTF8);

			StringBuilder rawTranslation = null;

			String[] urisInfos = SERVICE_QUERIES.get(this.serviceStringId);
			final String cookieURL = urisInfos[0];
			if (null != cookieURL) {

				final String serviceName = DictionaryOnlineActivity.this.getString(this.serviceStringId);
				HistoryDB historyDB = HistoryDB.getHistoryDB();

				CookieStore cookieStore = historyDB.getCookies(serviceName);
				if (cookieStore.getCookies().isEmpty()) {
					// Fetch cookie only if we don't have them
					HttpGet hg = new HttpGet(cookieURL);
					HttpResponse httpResponse = httpClient.execute(hg);

					int statusCode = httpResponse.getStatusLine().getStatusCode();
					if (HttpStatus.SC_OK == statusCode) {

						cookieStore = httpClient.getCookieStore();
						List<Cookie> cookies = cookieStore.getCookies();
						for (Cookie cookie : cookies) {
							Log.i(TAG, cookieURL + ": " + cookie.toString());
						}

						if (cookies.isEmpty()) {
							Log.w(TAG, "No cookie found");
						} else {
							historyDB.addCookies(serviceName, cookieStore);
						}

						httpResponse.getEntity().consumeContent();
					} else {
						rawTranslation = DictionaryOnlineActivity.getContent(httpResponse.getEntity().getContent());
						httpClient = null;
					}
				} else {
					// we have the cookies, no need to get them again
					httpClient.setCookieStore(cookieStore);
				}
			}

			if (null != httpClient) {

				String srcLang = DictionaryOnlineActivity.super.languageSource.langNewISO2;
				String dstLang = DictionaryOnlineActivity.super.languageDest.langNewISO2;

				String query = String.format(urisInfos[4], DictionaryOnlineActivity.this.editTextInput.getText(),
						srcLang, dstLang);
				URI uri = new URI(urisInfos[1], urisInfos[2], urisInfos[3], query, urisInfos[5]);

				// Log.d(TAG, "URL: " + uri.toString());

				HttpGet hg = new HttpGet(uri);
				HttpResponse httpResponse = httpClient.execute(hg);
				int statusCode = httpResponse.getStatusLine().getStatusCode();

				// Header[] headers = httpResponse.getAllHeaders();
				// Log.d(TAG, "headers:");
				// for (Header header : headers) {
				// Log.d(TAG, header.getName() + "=" + header.getValue() + "(" + header.toString() +
				// ")");
				// }

				if (HttpStatus.SC_OK == statusCode) {
					rawTranslation = DictionaryOnlineActivity.getContent(httpResponse.getEntity().getContent());
					// Log.d(TAG, "raw response: " + rawTranslation);
				} else {
					rawTranslation = DictionaryOnlineActivity.getContent(httpResponse.getEntity().getContent());
				}
			}

			return rawTranslation;
		}

		@Override
		protected void onPostExecute(StringBuilder result) {
			String term = DictionaryOnlineActivity.this.editTextInput.getText().toString();
			Language langDst = DictionaryOnlineActivity.super.languageDest;
			Language langSrc = DictionaryOnlineActivity.super.languageSource;

			try {
				OnlineResult onlineResult = new OnlineResult(this.serviceStringId, term, langSrc, langDst, result);

				DictionaryOnlineActivity.this.textViewResult.setText(Html.fromHtml(onlineResult.html));
				DictionaryOnlineActivity.this.addSearchResult(onlineResult);
			} catch (IOException ioe) {
				DictionaryOnlineActivity.this.textViewResult.setText(ioe.getMessage());
			}

			DictionaryOnlineActivity.this.setWidgetEnabled(true);
		}
	}

	@Override
	public void onClick(View v) {
		switch (v.getId()) {
		case R.id.dictionaryOnlineButtonLanguageDest:
			this.onClickButtonLanguageDest(v);
			break;
		case R.id.dictionaryOnlineButtonLanguageSource:
			this.onClickButtonLanguageSource(v);
			break;
		case R.id.dictionaryOnlineButtonLanguageSwap:
			this.onClickSwapLanguages(v);
			break;
		default:
			super.onClick(v);
		}
	}
}
