/*
 *  Copyright (C) 2012 Yvon TANGUY
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.models.dictionary;

import java.util.ArrayList;
import java.util.HashMap;

import org.vono.narau.R;
import org.vono.narau.common.Common;
import org.vono.narau.db.DatabaseType;
import org.vono.narau.db.DictionaryDB;
import org.vono.narau.preferences.Preferences;
import org.vono.narau.utils.Language;

import android.os.Parcel;
import android.os.Parcelable;

/**
 * <p>
 * entry (ent_seq, k_ele*, r_ele+, info?, sense+)
 * </p>
 * Entries consist of kanji elements, reading elements, general information and sense elements. Each
 * entry must have at least one reading element and one sense element. Others are optional.
 * 
 */
public class Entry implements Comparable<Entry>, Parcelable {

	public static final String HOST_INF = "dictionary.inf";

	private static final String[] SORT_ORDER = { "nf", "news", "ichi", "spec", "gai" };
	private static final int[] SORT_FACTOR = { 500, 12000, 12001, 12002, 12003 };

	private String htmlText;

	/**
	 * A unique numeric sequence number for each entry
	 */
	public int sequence;

	/**
	 * The kanji element, or in its absence, the reading element.
	 */
	public ArrayList<KanjiElement> k_ele;

	/**
	 * reading element
	 */
	public ArrayList<ReadingElement> r_ele;

	/**
	 * The sense element will record the translational equivalent of the Japanese word
	 */
	public ArrayList<Sense> senses;

	private int sortValue;

	/**
	 * Partial loading
	 */
	public boolean partial;

	public Entry(int id) {
		this.sequence = id;
		this.k_ele = null;
		this.r_ele = null;
		this.senses = null;
		this.htmlText = null;
		this.sortValue = Integer.MAX_VALUE;
		this.partial = false;
	}

	public void loadFull() {
		if (this.partial) {
			final Entry entry = DictionaryDB.loadEntry(this.sequence);
			this.partial = false;
			this.k_ele = entry.k_ele;
			this.r_ele = entry.r_ele;
			this.senses = entry.senses;
			this.sortValue = Integer.MAX_VALUE;
		}
	}

	public boolean haveLang(String lang) {
		if (null != this.senses) {
			for (Sense s : this.senses) {
				if (s.glosses.containsKey(lang)) {
					return true;
				}
			}
		}
		return false;
	}

	public String getFirstKanjiElement() {
		if (null != this.k_ele) {
			return this.k_ele.get(0).keb;
		}
		return null;
	}

	public String getFirstReadingElement() {
		if (null != this.r_ele) {
			return this.r_ele.get(0).reb;
		}
		return null;
	}

	public String getFirstGloss() {
		if (null != this.senses) {
			final ArrayList<Language> languages = Preferences.getLangsPriorities(DatabaseType.dictonary);
			for (Sense s : this.senses) {
				final HashMap<Language, String> glosses = s.glosses;
				for (Language language : languages) {
					if (glosses.containsKey(language)) {
						return glosses.get(language);
					}
				}
			}
		}
		return null;
	}

	public String getHTMLKanjiElems() {
		StringBuilder sb = new StringBuilder();
		if (null != this.k_ele) {
			final ArrayList<KanjiElement> listKE = this.k_ele;
			final int numKE = listKE.size();
			for (int i = 0; i < numKE; i++) {
				final KanjiElement ke = listKE.get(i);
				sb.append(ke.keb);
				sb.append(Entry.getInf(ke.ke_inf));
				sb.append("<br/>");
			}
		}

		return sb.toString();
	}

	public String getHTMLReadingElems() {
		StringBuilder sb = new StringBuilder();
		final ArrayList<ReadingElement> listRE = this.r_ele;
		final int numRE = listRE.size();

		for (int i = 0; i < numRE; i++) {
			final ReadingElement re = listRE.get(i);
			sb.append(re.reb);
			sb.append(Entry.getInf(re.re_inf));
			sb.append("<br/>");
		}

		return sb.toString();
	}

	public String getHTMLSenses() {
		StringBuilder sb = new StringBuilder();

		if (null != this.senses) {
			final ArrayList<Language> languages = Preferences.getLangsPriorities(DatabaseType.dictonary);
			final int numLangs = languages.size();

			final String senseString = Common.getString(R.string.dictionaryEntrySense);

			final ArrayList<Sense> listSenses = this.senses;
			final int numSenses = listSenses.size();

			for (int i = 0; i < numSenses; i++) {
				final Sense sense = listSenses.get(i);

				sb.append("<br/>&nbsp;&nbsp;&nbsp;<font color=\"#99FF99\">");
				sb.append(String.format(senseString, (i + 1)));

				String pos = "<br/>\n";
				if (null != sense.pos) {
					pos += "<i>";
					ArrayList<String> aPos = Common.splitString(sense.pos, Common.SLASH);
					int lenAPos = aPos.size();
					for (int p = 0; p < lenAPos; p++) {
						pos += "- " + Informations.getInfText(aPos.get(p)) + "<br/>\n";
					}
					pos += "</i>";
				}

				if (null != sense.stagr) {
					// // Only reading sense
					if (null == sense.stagk) {
						sb.append(":<br/>「" + sense.stagr + "」</font> " + pos);
					} else {
						sb.append(":<br/>「" + sense.stagk + " (" + sense.stagr + ")」</font> " + pos);
					}
				} else if (null != sense.stagk) {
					// Only kanji sense
					sb.append(":<br/>「" + sense.stagk + "」</font> " + pos);
				} else {
					sb.append(":</font> " + pos);
				}

				for (int l = 0; l < numLangs; l++) {
					final Language language = languages.get(l);

					// display the result in the right order
					if (sense.glosses.containsKey(language)) {
						String langName = language.displayName;
						sb.append("<font color=\"#FFAAAA\">");
						sb.append(langName);
						sb.append(":</font>&nbsp;");
						String[] aGloss = sense.glosses.get(language).split("/");
						for (int g = 0; g < aGloss.length; g++) {
							sb.append("<br/>&nbsp;&nbsp;&bull;&nbsp;");
							sb.append(aGloss[g]);
						}
						sb.append("</ul>");
						// sb.append();
						sb.append("<br/>\n");
					}
				}
			}
		}

		return sb.toString();
	}

	private static String getInf(ArrayList<String> inf) {
		if (null == inf || inf.size() == 0) {
			return Common.EMPTY_STRING;
		}

		final String infLink = "<a href=\"" + Common.SCHEME + "://" + HOST_INF + "/";

		StringBuilder sb = new StringBuilder(8);
		sb.append(" <i>" + infLink);
		sb.append(inf.get(0) + "\">" + inf.get(0) + "</a>");
		for (int i = 1; i < inf.size(); i++) {
			sb.append("," + infLink);
			sb.append(inf.get(i) + "\">" + inf.get(i) + "</a>");
		}
		sb.append("</i>");
		return sb.toString();
	}

	private String getJapaneseText() {
		StringBuilder sb = new StringBuilder();
		if (null != this.k_ele && this.k_ele.size() > 0) {
			final ArrayList<KanjiElement> listKE = this.k_ele;
			final int numKE = listKE.size();
			KanjiElement ke = listKE.get(0);
			sb.append(ke.keb);
			sb.append(Entry.getInf(ke.ke_inf));
			for (int i = 1; i < numKE; i++) {
				ke = listKE.get(i);
				sb.append(" ");
				sb.append(Common.SLASH);
				sb.append(" ");
				sb.append(ke.keb);
				sb.append(Entry.getInf(ke.ke_inf));
			}
		}

		boolean haveKanji = false;
		if (sb.length() > 0) {
			haveKanji = true;
		}

		final ArrayList<ReadingElement> listRE = this.r_ele;
		final int numRE = listRE.size();

		if (haveKanji)
			sb.append(" (");
		for (int i = 0; i < numRE; i++) {
			final ReadingElement re = listRE.get(i);
			if (i > 0)
				sb.append(Common.SLASH);
			sb.append(re.reb);
			sb.append(Entry.getInf(re.re_inf));
		}
		if (haveKanji)
			sb.append(")");
		return sb.toString();
	}

	public String toString(String lang) {
		StringBuilder sb = new StringBuilder();

		sb.append("<p>");
		sb.append(this.sequence);
		sb.append("<br/>\n");
		sb.append(this.getJapaneseText());

		if (null != this.senses) {
			for (int i = 0; i < this.senses.size(); i++) {
				Sense sense = this.senses.get(i);
				if (!sense.glosses.containsKey(lang)) {
					continue;
				}

				sb.append("<br/>&nbsp;&nbsp;&nbsp;<font color=\"#99FF99\">");
				sb.append(i + 1);

				if (null != sense.stagr) {
					// // Only reading sense
					if (null == sense.stagk) {
						sb.append("「" + sense.stagr + "」");
					} else {
						sb.append("「" + sense.stagk + " (" + sense.stagr + ")」");
					}
				} else if (null != sense.stagk) {
					// Only kanji sense
					sb.append("「" + sense.stagk + "」");
				}

				sb.append(":</font> ");
				sb.append(sense.glosses.get(lang));
			}
		}
		sb.append("</p>\n");

		return sb.toString();
	}

	@Override
	public String toString() {
		if (this.htmlText == null) {
			this.loadFull();

			StringBuilder sb = new StringBuilder();
			sb.append("<p>\n");
			sb.append(this.getJapaneseText());
			sb.append("<br/>\n");

			sb.append(this.getHTMLSenses());

			sb.append("</p>\n");
			this.htmlText = sb.toString();
		}

		return this.htmlText;
	}

	/**
	 * Sort for the KanjiToWords<br/>
	 * This is not a very good sorting method, but it seems to work. At least the result is better
	 * with than without :-)<br/>
	 * This is an ugly code!
	 */
	private void makeSortOrder() {

		final String[] sortOrder = Entry.SORT_ORDER;
		final int orderLen = sortOrder.length;
		final int[] sortFactor = SORT_FACTOR;

		int sortVal = Integer.MAX_VALUE;

		if (null != this.k_ele) {
			for (KanjiElement ke : this.k_ele) {

				final ArrayList<String> ke_pri = ke.ke_pri;
				if (null == ke_pri) {
					continue;
				}
				final int priLen = ke_pri.size();
				for (int j = 0; j < priLen; j++) {

					final String pri = ke.ke_pri.get(j);
					for (int so = 0; so < orderLen; so++) {

						final String soTxt = sortOrder[so];
						if (pri.startsWith(soTxt)) {
							final String valStr = pri.substring(soTxt.length());
							int val = sortFactor[so] * Integer.valueOf(valStr);
							if (val > sortVal) {
								sortVal = val;
							}
						}
					}
				}
			}
		}
		if (sortVal == Integer.MAX_VALUE) {
			sortVal = Integer.MAX_VALUE - 1;
		}
		this.sortValue = sortVal;
	}

	@Override
	public boolean equals(Object o) {
		if (o instanceof Entry) {
			return this.sequence == ((Entry) o).sequence;
		}
		return false;
	}

	@Override
	public int hashCode() {
		return this.sequence;
	}

	/**
	 * <p>
	 * Compare the frequencies of this Japanese entry.
	 * </p>
	 * {@inheritDoc}
	 */
	public int compareTo(Entry another) {

		if (Integer.MAX_VALUE == this.sortValue) {
			this.makeSortOrder();
		}
		if (Integer.MAX_VALUE == another.sortValue) {
			another.makeSortOrder();
		}

		return this.sortValue - another.sortValue;
	}

	// Parcelable:

	public int describeContents() {
		return 0;
	}

	public void writeToParcel(Parcel dest, int flags) {
		dest.writeInt(this.sequence);
	}

}
