/*
 *  Copyright (C) 2012 Yvon TANGUY
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.preferences;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.vono.narau.R;
import org.vono.narau.common.Common;
import org.vono.narau.db.DatabaseType;
import org.vono.narau.db.DictionaryDB;
import org.vono.narau.db.KanjiDB;
import org.vono.narau.utils.Language;

import android.content.Context;
import android.content.SharedPreferences;
import android.content.SharedPreferences.Editor;
import android.preference.PreferenceManager;
import android.util.SparseArray;
import android.widget.TextView;

/**
 * Manage store preferences, and expose them to other classes.
 * 
 */
public class Preferences {
	// private static final String TAG = Preferences.class.getSimpleName();
	private static final char LANG_SEP = ',';

	private static class PrefDefault {
		final int prefValueId;
		final String prefName;

		PrefDefault(int nameId, int valueId) {
			this.prefValueId = valueId;
			this.prefName = Common.getResourcesString(nameId);

			Preferences.defaultValues.put(nameId, this);
		}
	}

	private static SharedPreferences sharedPreferences = null;
	private static SparseArray<PrefDefault> defaultValues = null;
	private static final HashMap<DatabaseType, ArrayList<Language>> langsPriorities;
	private static float zoomValue;
	private static boolean firstRun = false;

	static {
		langsPriorities = new HashMap<DatabaseType, ArrayList<Language>>();
	}

	/**
	 * DictionaryDB must have been init before.
	 * 
	 * @param context
	 */
	@SuppressWarnings("unused")
	public static void init(Context context) {
		sharedPreferences = PreferenceManager.getDefaultSharedPreferences(context);
		Preferences.migrPref();

		Preferences.defaultValues = new SparseArray<PrefDefault>(8);

		new PrefDefault(R.string.prefNameVersion, R.integer.prefValueVersion);
		new PrefDefault(R.string.prefNameTextZoom, R.string.prefValueTextZoom);
		new PrefDefault(R.string.prefNameDictionaryMaxResult, R.integer.prefValueDictionaryMaxResult);
		new PrefDefault(R.string.prefNameDictionaryWordMatch, R.bool.prefValueDictionaryWordMatch);
		new PrefDefault(R.string.prefNameKanjiShowGrid, R.bool.prefValueKanjiShowGrid);
		new PrefDefault(R.string.prefNameKanjiShowStrokesNumber, R.bool.prefValueKanjiShowStrokesNumber);
		new PrefDefault(R.string.prefNameKanjiStrokesWidth, R.string.prefValueKanjiStrokeDefaultWidth);
		new PrefDefault(R.string.prefNameKanjiDrawSpeed, R.string.prefValueKanjiStrokeDefaultDrawSpeed);
		new PrefDefault(R.string.prefNameKanjiToWordsNWTL, R.integer.prefValueKanjiToWordsNWTL);
		new PrefDefault(R.string.prefNameHistoryDictionaryMax, R.integer.prefValueHistoryDictionaryMax);

		// Init the default text size
		// Not a very elegant method, but I don't know how to do better
		Preferences.zoomValue = Preferences.getFloat(R.string.prefNameTextZoom);
		if (Float.NaN == Preferences.zoomValue || 0f == Preferences.zoomValue) {
			Preferences.zoomValue = (new TextView(context)).getTextSize();
			Preferences.setFloat(R.string.prefNameTextZoom, Preferences.zoomValue);
		}
	}

	/**
	 * Used by Common, to free memory.
	 */
	public static void clean() {
		sharedPreferences = null;
		if (null != defaultValues) {
			defaultValues.clear();
			defaultValues = null;
		}
		langsPriorities.clear();
		Preferences.firstRun = false;
	}

	/**
	 * Clear all preferences
	 */
	public static void clear(Context context) {
		Preferences.sharedPreferences.edit().clear().commit();
		Preferences.langsPriorities.clear();
		Preferences.init(context);
	}

	public static boolean isFirstRun() {
		return Preferences.firstRun;
	}

	/**
	 * Update the preferences when there is some change (ex: key deleted, format change...).
	 */
	private static void migrPref() {
		final String prefName = Common.getResourcesString(R.string.prefNameVersion);
		final int prefValue = Common.getResourcesInteger(R.integer.prefValueVersion);
		int prefVersion = sharedPreferences.getInt(prefName, -1);

		if (prefValue == prefVersion) {
			// OK, good version
			return;
		}

		final Editor edit = sharedPreferences.edit();

		if (-1 == prefVersion) {
			// First time the application is launch
			edit.putInt(prefName, prefValue);
			edit.commit();
			Preferences.firstRun = true;
			return;
		}

		// Remove all keys
		final String oldDBPrefix = "db.creation.";
		Map<String, ?> mapAll = sharedPreferences.getAll();
		Set<String> keys = mapAll.keySet();
		for (String key : keys) {
			if (key.startsWith("kanjiItem")) {
				edit.remove(key);
			} else if (!key.contains(".")) {
				edit.remove(key);
			} else if (key.startsWith(oldDBPrefix)) {
				edit.remove(key);
			} else if ("dictionary.current_lang".equals(key)) {
				edit.remove(key);
			}
		}
		edit.putInt(prefName, prefValue);
		edit.commit();
	}

	public static SharedPreferences getSP() {
		return Preferences.sharedPreferences;
	}

	public static String getString(int prefNameId) {
		final PrefDefault prefDefault = defaultValues.get(prefNameId);
		final String prefName;
		final String prefValue;
		if (null == prefDefault) {
			prefName = Common.getResourcesString(prefNameId);
			prefValue = null;
		} else {
			prefName = prefDefault.prefName;
			prefValue = Common.getResourcesString(prefDefault.prefValueId);
		}

		return sharedPreferences.getString(prefName, prefValue);
	}

	public static int getInt(int prefNameId) {
		final PrefDefault prefDefault = defaultValues.get(prefNameId);
		if (null != prefDefault) {
			final int prefValue = Common.getResourcesInteger(prefDefault.prefValueId);
			return sharedPreferences.getInt(prefDefault.prefName, prefValue);
		}
		final String prefName = Common.getResourcesString(prefNameId);
		return sharedPreferences.getInt(prefName, 0);
	}

	public static void setInt(int prefNameId, int value) {
		final String prefName = Common.getResourcesString(prefNameId);
		sharedPreferences.edit().putInt(prefName, value).commit();
	}

	public static boolean getBoolean(int prefNameId) {
		final PrefDefault prefDefault = defaultValues.get(prefNameId);
		final boolean prefValue = Common.getResourcesBoolean(prefDefault.prefValueId);

		return sharedPreferences.getBoolean(prefDefault.prefName, prefValue);
	}

	public static float getFloat(int prefNameId) {
		if (null != defaultValues) {
			final PrefDefault prefDefault = defaultValues.get(prefNameId);
			final float prefValue = Common.getResourcesFloat(prefDefault.prefValueId);

			return sharedPreferences.getFloat(prefDefault.prefName, prefValue);
		}
		// This function is used by KanjiView to get the stroke width.
		// If the view is displayed in Eclipse by the Graphical Layout
		// it will crash without the null test.
		return Float.NaN;
	}

	/**
	 * 
	 * @param prefName
	 *            The preference name
	 * @return the value for this preference. If not found, return the default value as specified in
	 *         values.xml. If not found return Float.NaN.
	 */
	public static float getFloat(String prefName) {
		Float value = Float.NaN;
		if (null != sharedPreferences) {
			value = sharedPreferences.getFloat(prefName, Float.NaN);
			if (Float.NaN == value) {
				final int size = defaultValues.size();
				for (int i = 0; i < size; i++) {
					final PrefDefault prefDefault = defaultValues.valueAt(i);
					if (prefDefault.prefName.equals(prefName)) {
						value = Common.getResourcesFloat(prefDefault.prefValueId);
					}
				}
			}
		}

		return value;
	}

	public static void setString(int prefNameId, String value) {
		final String prefName = Common.getResourcesString(prefNameId);
		sharedPreferences.edit().putString(prefName, value).commit();
	}

	public static void setBoolean(String name, boolean value) {
		sharedPreferences.edit().putBoolean(name, value).commit();
	}

	public static void setBoolean(int prefNameId, boolean value) {
		final String prefName = Common.getResourcesString(prefNameId);
		sharedPreferences.edit().putBoolean(prefName, value).commit();
	}

	public static void setFloat(int prefNameId, float value) {
		final String prefName = Common.getResourcesString(prefNameId);
		sharedPreferences.edit().putFloat(prefName, value).commit();
	}

	public static void setFloat(String prefName, float value) {
		sharedPreferences.edit().putFloat(prefName, value).commit();
	}

	/**
	 * @return change the zoom value (*1.1), and return the new value.
	 */
	public static float zoomIn() {
		Preferences.zoomValue *= 1.1f;
		Preferences.setFloat(R.string.prefNameTextZoom, Preferences.zoomValue);
		return Preferences.zoomValue;
	}

	/**
	 * 
	 * @return change the zoom value (*0.9), and return the new value.
	 */
	public static float zoomOut() {
		Preferences.zoomValue *= 0.9f;
		Preferences.setFloat(R.string.prefNameTextZoom, Preferences.zoomValue);
		return Preferences.zoomValue;
	}

	/**
	 * 
	 * @return the current text zoom value.
	 */
	public static float getZoom() {
		return Preferences.zoomValue;
	}

	public static void setLangsPriorities(DatabaseType database, ArrayList<Language> values) {
		String strValue = Preferences.languageListToString(values, LANG_SEP);
		if (null != strValue) {
			String key = null;
			switch (database) {
			case dictonary:
				key = Common.getResourcesString(R.string.prefNameDictionaryLangPrio);
				break;
			case kanji:
				key = Common.getResourcesString(R.string.prefNameKanjiLangPrio);
				break;
			default:
				return;
			}

			langsPriorities.put(database, values);
			sharedPreferences.edit().putString(key, strValue).commit();
		}
	}

	public static ArrayList<Language> getLangsPriorities(DatabaseType database) {
		ArrayList<Language> languages = langsPriorities.get(database);
		if (null == languages) {

			String key = null;

			switch (database) {
			case dictonary:
				languages = DictionaryDB.getLangs();
				key = Common.getResourcesString(R.string.prefNameDictionaryLangPrio);
				break;
			case kanji:
			default:
				languages = KanjiDB.getLangs();
				key = Common.getResourcesString(R.string.prefNameKanjiLangPrio);
			}

			if (null != languages) {
				// null if no DB of that type

				String prefValueLangs = sharedPreferences.getString(key, null);
				if (null != prefValueLangs) {
					ArrayList<Language> prefLangs = Preferences.stringToLanguageList(prefValueLangs);

					// Add pref lang in right order
					// only if there is a DB with that lang
					for (Language language : prefLangs) {
						if (languages.contains(language)) {
							languages.remove(language);
							languages.add(languages.size(), language);
						}
					}
				}

				langsPriorities.put(database, languages);
			}

		}

		return languages;
	}

	private static ArrayList<Language> stringToLanguageList(String string) {
		final ArrayList<String> listLangs = Common.splitString(string, LANG_SEP);
		final int numLangs = listLangs.size();
		final ArrayList<Language> listLocales = new ArrayList<Language>(numLangs);

		for (int i = 0; i < numLangs; i++) {
			final String iso3 = listLangs.get(i);
			final Language language = Language.MAP_ISO3.get(iso3);
			listLocales.add(language);
		}

		return listLocales;
	}

	private static String languageListToString(ArrayList<Language> list, char sep) {
		String str = null;
		if (null != list) {
			int size = list.size();
			StringBuilder sb = new StringBuilder(20);
			for (int i = 0; i < size; i++) {
				String elem = list.get(i).langISO3;
				if (sb.length() > 0)
					sb.append(sep);
				sb.append(elem);

			}
			str = sb.toString();
		}
		return str;
	}
}
