/*
 *  Copyright (C) 2012 Yvon TANGUY
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.utils;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Locale;

import android.os.Parcel;
import android.os.Parcelable;

/**
 * Language informations extract from a Locale
 */
public class Language implements Comparable<Language>, CharSequence, Parcelable {

	public static final Language DEFAULT;
	public static final Language ENGLISH;
	public static final Language JAPANESE;

	public static final ArrayList<Language> LIST;
	public static final HashMap<String, Language> MAP_ISO3;

	/**
	 * OpenJDK locals.
	 */
	private static final String[][] LOCALES = { { "ukr", "uk", "Ukrainian" }, { "slk", "sk", "Slovak" },
			{ "cat", "ca", "Catalan" }, { "zho", "zh", "Chinese" }, { "deu", "de", "German" },
			{ "fra", "fr", "French" }, { "mlt", "mt", "Maltese" }, { "por", "pt", "Portuguese" },
			{ "hin", "hi", "Hindi" }, { "fin", "fi", "Finnish" }, { "heb", "iw", "Hebrew" }, { "ces", "cs", "Czech" },
			{ "swe", "sv", "Swedish" }, { "slv", "sl", "Slovenian" }, { "nld", "nl", "Dutch" },
			{ "srp", "sr", "Serbian" }, { "kor", "ko", "Korean" }, { "est", "et", "Estonian" },
			{ "ind", "in", "Indonesian" }, { "rus", "ru", "Russian" }, { "pol", "pl", "Polish" },
			{ "hrv", "hr", "Croatian" }, { "ara", "ar", "Arabic" }, { "msa", "ms", "Malay" }, { "gle", "ga", "Irish" },
			{ "eng", "en", "English" }, { "vie", "vi", "Vietnamese" }, { "spa", "es", "Spanish" },
			{ "lit", "lt", "Lithuanian" }, { "lav", "lv", "Latvian" }, { "hun", "hu", "Hungarian" },
			{ "isl", "is", "Icelandic" }, { "ron", "ro", "Romanian" }, { "ell", "el", "Greek" },
			{ "tur", "tr", "Turkish" }, { "bel", "be", "Belarusian" }, { "nor", "no", "Norwegian" },
			{ "bul", "bg", "Bulgarian" }, { "dan", "da", "Danish" }, { "sqi", "sq", "Albanian" },
			{ "ita", "it", "Italian" }, { "tha", "th", "Thai" }, { "mkd", "mk", "Macedonian" },
			{ "jpn", "ja", "Japanese" }, };

	static {
		// Static initalisation
		LIST = new ArrayList<Language>();
		MAP_ISO3 = new HashMap<String, Language>();

		Locale defaultLocale = Locale.getDefault();
		DEFAULT = new Language(defaultLocale);

		if (Locale.ENGLISH.getISO3Language().equals(defaultLocale.getISO3Language())) {
			ENGLISH = DEFAULT;
			LIST.add(DEFAULT);
			MAP_ISO3.put(DEFAULT.langISO3, DEFAULT);
		} else {
			ENGLISH = new Language(Locale.ENGLISH);
			LIST.add(DEFAULT);
			LIST.add(ENGLISH);
			MAP_ISO3.put(DEFAULT.langISO3, DEFAULT);
			MAP_ISO3.put(ENGLISH.langISO3, ENGLISH);
		}

		JAPANESE = new Language(Locale.JAPANESE);
		LIST.add(JAPANESE);
		MAP_ISO3.put(JAPANESE.langISO3, JAPANESE);

		Locale[] locales = Locale.getAvailableLocales();
		int num = locales.length;
		for (int i = 0; i < num; i++) {
			Locale locale = locales[i];
			if (!MAP_ISO3.containsKey(locale.getISO3Language())) {
				Language language = new Language(locale);
				MAP_ISO3.put(language.langISO3, language);
				LIST.add(language);
			}
		}

		// Problem with Cupcake which do not retrieve all locale with getAvailableLocales();
		// Build the locals based on OpenJDK list
		String[][] strLocales = LOCALES;
		num = strLocales.length;
		for (int i = 0; i < num; i++) {
			String[] locale = strLocales[i];
			if (!MAP_ISO3.containsKey(locale[0])) {
				Language language = new Language(locale[0], locale[1], Language.getNewLanguageCode(locale[1]),
						locale[2]);
				MAP_ISO3.put(language.langISO3, language);
				LIST.add(language);
			}
		}

		Collections.sort(LIST);
	}

	public final String langISO3;
	public final String langNewISO2;
	public final String langOldISO2;
	public final String displayName;

	private Language(Locale locale) {
		this.langISO3 = locale.getISO3Language().intern();
		this.langOldISO2 = locale.getLanguage();
		this.langNewISO2 = Language.getNewLanguageCode(this.langOldISO2).intern();
		this.displayName = locale.getDisplayLanguage().intern();
	}

	public Language(String iso3, String oldIso2, String newIso2, String name) {
		this.langISO3 = iso3.intern();
		this.langOldISO2 = oldIso2.intern();
		this.langNewISO2 = newIso2.intern();
		this.displayName = name.intern();
	}

	public int compareTo(Language another) {

		if (this == DEFAULT || DEFAULT.equals(this)) {
			return -1;
		} else if (another == DEFAULT || DEFAULT.equals(another)) {
			return 1;
		} else if (this == ENGLISH || ENGLISH.equals(this)) {
			return -1;
		} else if (another == ENGLISH || ENGLISH.equals(another)) {
			return 1;
		}

		return this.displayName.compareTo(another.displayName);
	}

	@Override
	public String toString() {
		return this.displayName;
	}

	@Override
	public int hashCode() {
		return this.langISO3.hashCode() + this.displayName.hashCode();
	}

	@Override
	public boolean equals(Object o) {
		if (this == o) {
			return true;
		}

		if (o instanceof Language) {
			return this.langISO3.equals(((Language) o).langISO3);
		}

		return false;
	}

	/**
	 * Map obsolete language codes to the new language codes so the correct translation will be
	 * used.
	 * 
	 * @param languageCode
	 *            old language code used by Java
	 * @return the new language code if needed, or the current language code if no replacement was
	 *         made.
	 * @see Locale
	 */
	private static String getNewLanguageCode(String languageCode) {
		if (languageCode.equals("iw")) {
			return "he";
		} else if (languageCode.equals("in")) {
			return "id";
		} else if (languageCode.equals("ji")) {
			return "yi";
		}

		return languageCode;
	}

	public int length() {
		return this.displayName.length();
	}

	public char charAt(int index) {
		return this.displayName.charAt(index);
	}

	public CharSequence subSequence(int start, int end) {
		return this.displayName.subSequence(start, end);
	}

	public int describeContents() {
		// Parcelable
		return 0;
	}

	public void writeToParcel(Parcel dest, int flags) {
		dest.writeString(this.langISO3);

	}

	public static final Parcelable.Creator<Language> CREATOR = new Parcelable.Creator<Language>() {
		public Language createFromParcel(Parcel in) {
			String iso3 = in.readString();
			return MAP_ISO3.get(iso3);
		}

		public Language[] newArray(int size) {
			return new Language[size];
		}
	};
}
