/*
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.dictdb;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.regex.Pattern;

import org.vono.narau.Common;
import org.vono.narau.Log;
import org.vono.narau.dictdb.Dictionary.WritingType;
import org.vono.narau.dictdb.entry.Entry;
import org.vono.narau.dictdb.entry.KanjiElement;
import org.vono.narau.dictdb.entry.ReadingElement;
import org.vono.narau.dictdb.entry.Sense;
import org.vono.narau.dictdb.entry.SenseGloss;

/**
 * Read an EDict file.<br/>
 * TODO: EDict2 file support.
 */
public class EDictReader {
	private static final String TAG = EDictReader.class.getSimpleName();

	private static final char CHAR_LINEFEED = '\n';
	private static final char CHAR_CARRIAGERETURN = '\r';
	public static final char CHAR_SPACE = ' ';
	public static final char CHAR_SLASH = '/';

	private String error;
	private final Dictionary dictionary;

	public EDictReader(Dictionary dictionary) {
		this.dictionary = dictionary;
	}

	public String getError() {
		return this.error;
	}

	public void read(String lang, File edictFile) throws IOException {

		final FileReader fileReader = new FileReader(edictFile);
		try {
			Log.i(TAG, "reading " + edictFile.getAbsolutePath());
			this.readDict(lang, fileReader);
		} finally {
			fileReader.close();
		}
	}

	private void readDict(String lang, FileReader fr) throws IOException {
		char[] buffer = new char[1024];
		int strLen = 0;
		int line = 0;
		int subStrNum = 0;

		EDictEntry editcEntry = new EDictEntry(lang);

		boolean highSurrogate = false;
		boolean supplementaryCodePoint = false;
		boolean badUnicode = false;
		int maxLen = 0;
		int numNew = 0;

		for (int c = fr.read(); c != -1; c = fr.read()) {
			char cc = (char) c;
			if (Character.isHighSurrogate(cc)) {
				highSurrogate = true;
			}

			if (!Character.isDefined(c)) {
				badUnicode = true;
				continue;
			}

			if (Character.isSupplementaryCodePoint(c)) {
				supplementaryCodePoint = true;
			}
			switch (c) {
			case CHAR_CARRIAGERETURN:
				// ignore
				break;
			case CHAR_LINEFEED:
				line++;

				editcEntry.addGloss(buffer, strLen);
				if (editcEntry.isValid()) {
					WritingType writingType = WritingType.kanji;
					String jpn = editcEntry.kanji;
					if (null == jpn) {
						writingType = WritingType.kana;
						jpn = editcEntry.kana;
					}
					// It is slow to search for every entries for a common
					// match,but it is not really a problem since this program
					// is not executed very often.
					Entry entry = this.dictionary.findEntryFromJpn(jpn, writingType);
					if (null != entry) {
						editcEntry.mergeToEntry(entry);
					} else {
						Log.i(TAG, "Line " + line + ": New entry created: " + editcEntry.toString());
						entry = editcEntry.toEntry(this.dictionary);
						this.dictionary.addEntry(entry);
						numNew++;
					}
				} else {
					Log.i(TAG, "Line " + line + ": invalid line");
				}

				// Faster to clean than create a new one ?
				editcEntry.clear();
				strLen = 0;
				subStrNum = 0;
				break;
			case CHAR_SLASH:
				if (0 == subStrNum) {
					editcEntry.setKanjiKana(buffer, strLen);
				} else {
					editcEntry.addGloss(buffer, strLen);
				}
				strLen = 0;
				subStrNum++;
				break;
			case CHAR_SPACE:
				if (strLen == 0) {
					break;
				}
				//$FALL-THROUGH$
			default:
				if (strLen < buffer.length) {
					buffer[strLen] = (char) c;
				} else if (maxLen < strLen) {
					maxLen = strLen;
				}
				strLen++;

			}
		}
		buffer = null;

		if (highSurrogate) {
			Log.w(TAG, "HighSurrogate char present");
		}
		if (supplementaryCodePoint) {
			Log.w(TAG, "SupplementaryCodePoint char present");
		}
		if (badUnicode) {
			Log.w(TAG, "Found unicode char not defined");
		}
		if (maxLen > 0) {
			Log.w(TAG, "Max line len reach: " + maxLen);
			Log.w(TAG, "you should modify the line: \"char[] buffer = new char[1024]\" and change 1024 to " + maxLen);
		}
		if (numNew > 0) {
			Log.i(TAG, "Number of new entries created: " + numNew);
		}
	}

	/**
	 * 
	 * <p>
	 * The EDICT file is in a relatively simple format based on the text data
	 * file of the SKK input-method. Each entry is in the form:
	 * <ul>
	 * <li>KANJI [KANA] /(general information) gloss/gloss/.../</li>
	 * <li>KANA /(general information) gloss/gloss/.../</li>
	 * </ul>
	 * <i>Source: <a
	 * href="http://www.csse.monash.edu.au/~jwb/edict_doc.html">The Electronic
	 * Dictionary Research and Development Group</a></i><br/>
	 * Copyright (C) 2010 The Electronic Dictionary Research and Development
	 * Group
	 * </p>
	 */
	private static class EDictEntry {
		final static char CHAR_OPEN_SQUAREBRACKET = '[';
		final static char CHAR_CLOSE_SQUAREBRACKET = ']';
		final static char CHAR_OPEN_PARENTHESE = '(';
		final static char CHAR_CLOSE_PARENTHESE = ')';
		final static String CHAR_COMMA = ",";

		String edictString;
		String kanji;
		String kana;

		/**
		 * default translation (to take kanjis/kana from)
		 */
		final String lang;
		final ArrayList<Sense> senses;

		public EDictEntry(String lang) {
			this.lang = lang;
			this.senses = new ArrayList<Sense>();
			this.edictString = null;
			this.kana = null;
			this.kanji = null;
		}

		/**
		 * Function used when reading EDIT v1 file format
		 * 
		 * @param entry
		 *            the same entry
		 * @return true if the merge is OK
		 * 
		 */
		void mergeToEntry(Entry entry) {
			final ArrayList<Sense> eSenses = entry.senses;
			final int numESenses = eSenses.size();

			final int commonMaxNumSenses = Math.min(numESenses, this.senses.size());

			int count = 0;
			for (; count < commonMaxNumSenses; count++) {
				Sense ourSense = this.senses.get(count);
				Sense entSense = eSenses.get(count);

				// Merge glosses
				for (SenseGloss sg : ourSense.glosses) {
					if (!entSense.glosses.contains(sg)) {
						entSense.glosses.add(sg);
					}
				}
				// Merge pos
				if (null != ourSense.pos && !ourSense.pos.isEmpty()) {
					if (null == entSense.pos || entSense.pos.isEmpty()) {
						entSense.pos = ourSense.pos;
					} else {
						for (String s : ourSense.pos) {
							if (!entSense.pos.contains(s)) {
								entSense.pos.add(s);
							}
						}
					}
				}
			}

			for (; count < this.senses.size(); count++) {
				entry.senses.add(this.senses.get(count));
			}
		}

		Entry toEntry(Dictionary dictionary) {
			final int id = dictionary.getMaxEntryId() + 1;
			Entry entry = new Entry(id);

			if (null != this.kanji) {
				KanjiElement kElem = new KanjiElement();
				kElem.keb = this.kanji;
				entry.k_ele.add(kElem);
			}
			ReadingElement rElem = new ReadingElement();
			rElem.reb = this.kana;
			entry.r_ele.add(rElem);

			entry.senses.addAll(this.senses);

			return entry;
		}

		/**
		 * Reset this entry, to his original state
		 * 
		 */
		public void clear() {
			this.edictString = null;
			this.kanji = null;
			this.kana = null;

			this.senses.clear();
		}

		public void addGloss(char[] buffer, int len) {

			if (len == 0) {
				return;
			}

			len = len > buffer.length ? buffer.length : len;
			String originalGloss = new String(buffer, 0, len);
			String text = originalGloss.trim();

			if (text.isEmpty()) {
				return;
			}

			// Extract what is in the parentheses (if exists)
			int senseNum = 0;
			boolean haveSenseNum = false;
			HashSet<String> pos = new HashSet<String>();

			String insideParenthese = null;
			do {
				insideParenthese = null;
				if (text.charAt(0) == CHAR_OPEN_PARENTHESE) {
					int end = text.indexOf(CHAR_CLOSE_PARENTHESE);
					if (end > 1) {
						insideParenthese = text.substring(1, end).trim();

						if (text.length() > end + 1) {
							text = text.substring(end + 1).trim();
						} else {
							text = Common.EMPTY_STRING;
						}
					}
				}

				if (null != insideParenthese) {
					if (Pattern.matches("[0-9]+", insideParenthese)) {
						if (haveSenseNum) {
							Log.w(TAG, "double sense number for: " + originalGloss);
						}
						// test if it is a number (gloss sense)
						senseNum = Integer.parseInt(insideParenthese) - 1;
						if (senseNum < 0)
							// not possible ?
							senseNum = 0;

						haveSenseNum = true;
					} else {
						// general information
						final String[] sa = insideParenthese.split(CHAR_COMMA);
						for (int saIdx = 0; saIdx < sa.length; saIdx++) {
							String gi = sa[saIdx].trim().intern();

							if (gi.isEmpty() || "P".equals(gi)) {
								// ignore
								/*
								  http://www.csse.monash.edu.au/~jwb/wwwjdicinf.html#dicfil_tag <br/>
								  <ul>
								  <li>[Q] I understand the (P) on some entries means it is a common
								  word. Why is the (P) sometimes attached to the kanji or reading? <br/>
								  [A] Usually the (P) is at the end of the entry. When there is more
								  than one kanji headword and/or reading, the (P) is placed near the
								  headword or reading which is the common one.</li>
								  <li>[Q] Some of the words marked (P) are actually not very common,
								  and also there are some common words not marked. Why is this? <br/>
								  [A] The allocation of those (P) markers is based on a number of
								  sources, and inevitably has some problems. (For more information
								  about the markers, see the Word Priority Marking section in the
								  JMdict/EDICT documentation.) If you see any that are dubious, or see
								  entries that you think deserve a marker, please use the amendment
								  form to suggest a change.</li>
								  <ul>*/
							} else if (Sense.SET_POS.contains(gi)) {
								pos.add(gi);
							} else {
								text = text + " (" + gi + ")";
								// Log.w(TAG, "wrong gen info: " + gi + " -> " +
								// text);
							}
						}
					}
				}
			} while (!(insideParenthese == null || text.isEmpty()));

			Sense sense = null;
			if (this.senses.size() > senseNum) {
				sense = this.senses.get(senseNum);
			} else if (this.senses.size() == senseNum) {
				sense = new Sense();
				this.senses.add(sense);
			} else {
				Log.w(TAG, "We do not have enough sense: " + originalGloss);
				while (this.senses.size() <= (senseNum + 1)) {
					sense = new Sense();
					this.senses.add(sense);
				}
			}

			if (!text.isEmpty()) {
				SenseGloss gloss = new SenseGloss();
				gloss.text = text;
				gloss.lang = this.lang;
				sense.glosses.add(gloss);
				if (!pos.isEmpty()) {
					if (null == sense.pos || sense.pos.isEmpty()) {
						sense.pos = pos;
					} else {
						sense.pos.addAll(pos);
					}
				}
			}

		}

		public void setKanjiKana(char[] buffer, int length) {

			// this function is not fault tolerant for speed reason
			int kanaPos = 0;
			final int len = length > buffer.length ? buffer.length : length;

			if (len == 0) {
				return;
			}

			if (buffer[0] == '？') {
				return;
			}

			// skip the first 'word' which is a line number
			int startIndex = 0;
			for (startIndex = 0; startIndex < len; startIndex++) {
				final char c = buffer[startIndex];
				if (!(c >= '0' && c <= '9')) {
					break;
				}
			}

			for (int i = startIndex; i < len; i++) {
				final char c = buffer[i];

				switch (c) {
				case CHAR_OPEN_SQUAREBRACKET:
					if (i > 1) {
						this.kanji = new String(buffer, 0, i - 1).trim();
					}
					kanaPos = i + 1;
					break;
				case CHAR_CLOSE_SQUAREBRACKET:
					this.kana = new String(buffer, kanaPos, i - kanaPos).trim();
					i = len;
					break;
				}
			}
			if (null == this.kana && startIndex < len) {
				this.kana = new String(buffer, startIndex, len).trim();
			}
		}

		/**
		 * test if kana is set and this entry have at least one translation.
		 * 
		 * @return true if valid
		 */
		public boolean isValid() {

			if (null != this.kana && this.senses.size() > 0) {
				Sense sense = this.senses.get(0);
				return !sense.glosses.isEmpty();
			}

			return false;
		}

		/**
		 * Return the translations of this entry.
		 * 
		 * @return the translation or null if not present.
		 */
		public String getGlosses() {
			StringBuilder sb = new StringBuilder();

			final ArrayList<Sense> lSenses = this.senses;
			final int numSenses = lSenses.size();
			for (int sIdx = 0; sIdx < numSenses; sIdx++) {
				final ArrayList<SenseGloss> lGlosses = lSenses.get(sIdx).glosses;
				final int numGlosses = lGlosses.size();
				for (int gIdx = 0; gIdx < numGlosses; gIdx++) {
					sb.append(EDictReader.CHAR_SLASH);
					sb.append(CHAR_OPEN_PARENTHESE);
					sb.append(Integer.toString(sIdx + 1));
					sb.append(CHAR_CLOSE_PARENTHESE);
					sb.append(' ');
					sb.append(lGlosses.get(gIdx));
				}
			}

			return sb.toString();
		}

		/**
		 * Edict2 line format.
		 */
		@Override
		public String toString() {
			if (null == this.edictString) {
				if (this.isValid()) {

					StringBuilder sb = new StringBuilder();
					if (null != this.kanji) {
						sb.append(this.kanji);
						sb.append(EDictReader.CHAR_SPACE);
						sb.append(CHAR_OPEN_SQUAREBRACKET);
						sb.append(this.kana);
						sb.append(CHAR_CLOSE_SQUAREBRACKET);
					} else {
						sb.append(this.kana);
					}

					sb.append(this.getGlosses());
					this.edictString = sb.toString();
				} else {
					this.edictString = Common.EMPTY_STRING;
				}
			}
			return this.edictString;
		}
	}
}
