/**
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.dictdb.entry;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.vono.narau.Log;

/**
 * <p>
 * entry (ent_seq, k_ele*, r_ele+, info?, sense+)
 * </p>
 * Entries consist of kanji elements, reading elements, general information and
 * sense elements. Each entry must have at least one reading element and one
 * sense element. Others are optional.
 * 
 */
public class Entry {
	private static final String TAG = Entry.class.getSimpleName();

	/**
	 * A unique numeric sequence number for each entry
	 */
	public int sequence;

	/**
	 * The kanji element, or in its absence, the reading element.
	 */
	public ArrayList<KanjiElement> k_ele;

	/**
	 * reading element
	 */
	public ArrayList<ReadingElement> r_ele;

	/**
	 * general information
	 */
	public Info info;

	/**
	 * sense elements
	 */
	public ArrayList<Sense> senses;

	public Entry(int seq) {
		this.sequence = seq;
		this.info = null;
		this.k_ele = new ArrayList<KanjiElement>();
		this.r_ele = new ArrayList<ReadingElement>(1);
		this.senses = new ArrayList<Sense>(1);
	}

	private static final Pattern patternNumber = Pattern.compile(".*\\([0-9]+\\).*");
	private static final Pattern patternParentheses = Pattern.compile("^(\\([a-zA-Z0-9]+\\))+");
	private static final HashMap<String, ArrayList<SenseGloss>> sMapGlosses = new HashMap<String, ArrayList<SenseGloss>>();

	/**
	 * Parse the sense to look for (&lt;number&gt;) in glosses, and rearrange
	 * them in the senses. Extract also the general information if found.
	 * 
	 * @return the number of original sense change
	 */
	public int fixGlosses() {
		final ArrayList<Sense> lSenses = this.senses;
		final int numSenses = lSenses.size();

		int numFix = 0;

		// Fix German/Dutch glosses which are always on the last sense
		numFix += this.fixLastGlossesLangOnly("deu");
		numFix += this.fixLastGlossesLangOnly("nld");

		HashMap<String, ArrayList<SenseGloss>> mapGlosses = Entry.sMapGlosses;
		mapGlosses.clear();

		// Search all glosses before
		for (int i = 0; i < numSenses; i++) {
			final Sense sense = lSenses.get(i);
			boolean found = false;
			final int numGlosses = sense.glosses.size();
			for (int j = 0; j < numGlosses; j++) {
				SenseGloss gloss = sense.glosses.get(j);
				gloss.text = gloss.text.trim();

				if (('(' == gloss.text.charAt(0) && patternNumber.matcher(gloss.text).matches())
						&& !mapGlosses.containsKey(gloss.lang)) {
					mapGlosses.put(gloss.lang, new ArrayList<SenseGloss>());
					found = true;
				}
			}
			if (found)
				numFix++;
		}

		if (!mapGlosses.isEmpty()) {

			// Log.i(TAG, "Found error in " + this.sequence + " " +
			// mapGlosses.keySet().toString());

			// Fetch all glosses for the langs with a pb
			for (int i = 0; i < numSenses; i++) {
				final Sense sense = lSenses.get(i);
				int numGlosses = sense.glosses.size();
				for (int j = 0; j < numGlosses;) {
					SenseGloss gloss = sense.glosses.get(j);
					if (mapGlosses.containsKey(gloss.lang)) {
						mapGlosses.get(gloss.lang).add(gloss);
						sense.glosses.remove(j);
						numGlosses--;
					} else {
						j++;
					}
				}
			}

			final HashSet<String> listPos = new HashSet<String>();
			// put the glosses in the right sense
			for (ArrayList<SenseGloss> glosses : mapGlosses.values()) {
				int numGlosses = glosses.size();
				for (int i = 0; i < numGlosses; i++) {

					listPos.clear();

					SenseGloss gloss = glosses.get(i);

					gloss.text = gloss.text.trim();

					int sensePosition = 0;

					Matcher matcher = patternParentheses.matcher(gloss.text);
					while (matcher.find()) {

						String first = matcher.group().trim();
						String insideParentheses = first.substring(1, first.length() - 1).intern();
						gloss.text = gloss.text.substring(first.length()).trim();

						// Check if it is a valid pos
						if (patternNumber.matcher(first).matches()) {
							if (0 != sensePosition) {
								Log.w(TAG, "double sense number for: " + this.sequence + " " + gloss.lang + ":"
										+ gloss.text);
							}
							sensePosition = Integer.parseInt(insideParentheses) - 1;
						} else if (Sense.SET_POS.contains(insideParentheses)) {
							listPos.add(insideParentheses);
						} else if (!insideParentheses.isEmpty()) {
							gloss.text = gloss.text + " (" + insideParentheses + ")";
							// Log.w(TAG, "wrong gen info: " + insideParentheses
							// + " -> " + gloss.text);
						}

						matcher = patternParentheses.matcher(gloss.text);
					}

					// merge gloss to sense
					Sense sense = null;
					if (this.senses.size() > sensePosition) {
						sense = this.senses.get(sensePosition);
					} else if (this.senses.size() == sensePosition) {
						sense = new Sense();
						this.senses.add(sense);
					} else {
						Log.w(TAG, "We do not have enough sense: " + gloss.toString());
						while (this.senses.size() <= (sensePosition + 1)) {
							sense = new Sense();
							this.senses.add(sense);
						}
					}
					if (!sense.glosses.contains(gloss)) {
						sense.glosses.add(gloss);
					}
					// merge pos: Part-of-speech information
					if (!listPos.isEmpty()) {
						if (null == sense.pos || sense.pos.isEmpty()) {
							sense.pos = listPos;
						} else {
							sense.pos.addAll(listPos);
						}
					}
				}
			}

			mapGlosses.clear();
		} // else, nothing to do

		// cleaning:
		this.deleteEmptySense();

		return numFix;
	}

	private int fixLastGlossesLangOnly(String lang) {
		final ArrayList<Sense> lSenses = this.senses;
		final int numSenses = lSenses.size();

		int numFix = 0;

		if (1 == numSenses) {
			final Sense firstSense = lSenses.get(0);
			final int numGlosses = firstSense.glosses.size();
			for (int i = 0; i < numGlosses; i++) {
				SenseGloss gloss = firstSense.glosses.get(i);
				gloss.text = gloss.text.trim();
				Matcher matcher = patternParentheses.matcher(gloss.text);
				while (null != matcher && matcher.find()) {
					String first = matcher.group().trim();
					String insideParentheses = first.substring(1, first.length() - 1).intern();

					if (Sense.SET_POS.contains(insideParentheses)) {
						gloss.text = gloss.text.substring(first.length()).trim();
						firstSense.pos.add(insideParentheses);
						matcher = patternParentheses.matcher(gloss.text);
					} else {
						matcher = null;
					}
				}
			}

			return 0;
		}

		final boolean[] haveLang = new boolean[numSenses];
		final ArrayList<SenseGloss> lastLangGlosses = new ArrayList<SenseGloss>();

		for (int i = 0; i < numSenses; i++) {
			haveLang[i] = false;
			final Sense sense = lSenses.get(i);
			int numGlosses = sense.glosses.size();
			for (int j = 0; j < numGlosses; j++) {
				SenseGloss gloss = sense.glosses.get(j);
				if (lang.equals(gloss.lang)) {
					haveLang[i] = true;

					if (i == (numSenses - 1)) {
						lastLangGlosses.add(gloss);
						numGlosses--;
						sense.glosses.remove(j);
						j--;
					} else {
						j = numGlosses;
					}
				}
			}
		}

		boolean haveLangBeforeLast = false;
		for (int i = 0; i < numSenses - 1; i++) {
			haveLangBeforeLast = haveLangBeforeLast || haveLang[i];
		}

		if (!haveLangBeforeLast && haveLang[numSenses - 1]) {
			// Move the German glosses to the first positions
			numFix++;
			// Fix general info
			final Sense firstSense = lSenses.get(0);
			final int numGlosses = lastLangGlosses.size();
			for (int i = 0; i < numGlosses; i++) {
				SenseGloss gloss = lastLangGlosses.get(i);
				gloss.text = gloss.text.trim();
				Matcher matcher = patternParentheses.matcher(gloss.text);
				while (null != matcher && matcher.find()) {
					String first = matcher.group().trim();
					String insideParentheses = first.substring(1, first.length() - 1).intern();

					if (Sense.SET_POS.contains(insideParentheses)) {
						gloss.text = gloss.text.substring(first.length()).trim();
						firstSense.pos.add(insideParentheses);
						matcher = patternParentheses.matcher(gloss.text);
					} else {
						matcher = null;
					}
				}
			}
			firstSense.glosses.addAll(lastLangGlosses);
		} else {
			lSenses.get(numSenses - 1).glosses.addAll(lastLangGlosses);
		}
		lastLangGlosses.clear();

		return numFix;
	}

	private int deleteEmptySense() {
		final ArrayList<Sense> lSenses = this.senses;
		int numSenses = lSenses.size();
		int numDelete = 0;
		for (int i = 0; i < numSenses;) {
			Sense s = lSenses.get(i);
			if (s.glosses.isEmpty()) {
				numSenses--;
				numDelete++;
				lSenses.remove(i);
			} else {
				i++;
			}
		}

		return numDelete;
	}

	@Override
	public boolean equals(Object obj) {
		if (obj instanceof Entry) {
			final Entry e = (Entry) obj;
			return this.sequence == e.sequence;
		}
		return false;
	}

	@Override
	public String toString() {
		final StringBuilder sb = new StringBuilder();

		sb.append(this.sequence);
		sb.append(' ');

		final boolean haveKE = this.k_ele.size() > 0;

		if (haveKE) {
			for (KanjiElement ke : this.k_ele) {
				sb.append(ke);
				sb.append(";");
			}
			sb.append(" [");
		}
		for (ReadingElement re : this.r_ele) {
			sb.append(re);
			sb.append(";");
		}
		if (haveKE) {
			sb.append(']');
		}

		for (Sense s : this.senses) {
			sb.append('/');
			sb.append(s);
		}

		return sb.toString();
	}

}
