/**
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.dictdb.entry;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;

/**
 * <p>
 * sense (stagk*, stagr*, pos*, xref*, ant*, field*, misc*, s_inf*, lsource*,
 * dial*, gloss*, example*)
 * </p>
 * The sense element will record the translational equivalent of the Japanese
 * word, plus other related information. Where there are several distinctly
 * different meanings of the word, multiple sense elements will be employed.
 * 
 */
public class Sense {

	public static final String[][] POS = { { "MA", "martial arts term" },
			{ "X", "rude or X-rated term (not displayed in educational software)" }, { "abbr", "abbreviation" },
			{ "adj-i", "adjective (keiyoushi)" }, { "adj-na", "adjectival nouns or quasi-adjectives (keiyodoshi)" },
			{ "adj-no", "nouns which may take the genitive case particle `no'" },
			{ "adj-pn", "pre-noun adjectival (rentaishi)" }, { "adj-t", "`taru' adjective" },
			{ "adj-f", "noun or verb acting prenominally" },
			{ "adj", "former adjective classification (being removed)" }, { "adv", "adverb (fukushi)" },
			{ "adv-to", "adverb taking the `to' particle" }, { "anat", "anatomical thingos" }, { "arch", "archaism" },
			{ "ateji", "ateji (phonetic) reading" }, { "aux", "auxiliary" }, { "aux-v", "auxiliary verb" },
			{ "aux-adj", "auxiliary adjective" }, { "Buddh", "Buddhist term" }, { "chem", "chemistry term" },
			{ "chn", "children's language" }, { "col", "colloquialism" }, { "comp", "computer terminology" },
			{ "conj", "conjunction" }, { "ctr", "counter" }, { "derog", "derogatory" }, { "eK", "exclusively kanji" },
			{ "engr", "engineering term" }, { "ek", "exclusively kana" },
			{ "exp", "Expressions (phrases, clauses, etc.)" }, { "fam", "familiar language" },
			{ "fem", "female term or language" }, { "finc", "Finnish word" }, { "food", "food term" },
			{ "geom", "geometry term" },
			{ "gikun", "gikun (meaning as reading)  or jukujikun (special kanji reading)" },
			{ "hon", "honorific or respectful (sonkeigo) language" }, { "hum", "humble (kenjougo) language" },
			{ "iK", "word containing irregular kanji usage" }, { "id", "idiomatic expression" },
			{ "ik", "word containing irregular kana usage" }, { "int", "interjection (kandoushi)" },
			{ "io", "irregular okurigana usage" }, { "iv", "irregular verb" }, { "ling", "linguistics terminology" },
			{ "m-sl", "manga slang" }, { "male", "male term or language" }, { "male-sl", "male slang" },
			{ "math", "mathematics" }, { "mil", "military" }, { "n", "noun (common) (futsuumeishi)" },
			{ "n-adv", "adverbial noun (fukushitekimeishi)" }, { "n-suf", "noun, used as a suffix" },
			{ "n-pref", "noun, used as a prefix" }, { "n-t", "noun (temporal) (jisoumeishi)" }, { "num", "numeric" },
			{ "oik", "old or irregular kana form" }, { "oK", "word containing out-dated kanji" },
			{ "obs", "obsolete term" }, { "obsc", "obscure term" }, { "ok", "out-dated or obsolete kana usage" },
			{ "on-mim", "onomatopoeic or mimetic word" }, { "pn", "pronoun" }, { "poet", "poetical term" },
			{ "pol", "polite (teineigo) language" }, { "pref", "prefix" }, { "proverb", "proverb" },
			{ "prt", "particle" }, { "physics", "physics terminology" }, { "rare", "rare" }, { "sens", "sensitive" },
			{ "sl", "slang" }, { "suf", "suffix" }, { "uK", "word usually written using kanji alone" },
			{ "uk", "word usually written using kana alone" }, { "v1", "Ichidan verb" },
			{ "v2a-s", "Nidan verb with 'u' ending (archaic)" }, { "v4h", "Yodan verb with `hu/fu' ending (archaic)" },
			{ "v4r", "Yodan verb with `ru' ending (archaic)" }, { "v5", "Godan verb (not completely classified)" },
			{ "v5aru", "Godan verb - -aru special class" }, { "v5b", "Godan verb with `bu' ending" },
			{ "v5g", "Godan verb with `gu' ending" }, { "v5k", "Godan verb with `ku' ending" },
			{ "v5k-s", "Godan verb - Iku/Yuku special class" }, { "v5m", "Godan verb with `mu' ending" },
			{ "v5n", "Godan verb with `nu' ending" }, { "v5r", "Godan verb with `ru' ending" },
			{ "v5r-i", "Godan verb with `ru' ending (irregular verb)" }, { "v5s", "Godan verb with `su' ending" },
			{ "v5t", "Godan verb with `tsu' ending" }, { "v5u", "Godan verb with `u' ending" },
			{ "v5u-s", "Godan verb with `u' ending (special class)" },
			{ "v5uru", "Godan verb - Uru old class verb (old form of Eru)" },
			{ "vz", "Ichidan verb - zuru verb (alternative form of -jiru verbs)" }, { "vi", "intransitive verb" },
			{ "vk", "Kuru verb - special class" }, { "vn", "irregular nu verb" },
			{ "vr", "irregular ru verb, plain form ends with -ri" },
			{ "vs", "noun or participle which takes the aux. verb suru" },
			{ "vs-c", "su verb - precursor to the modern suru" }, { "vs-s", "suru verb - special class" },
			{ "vs-i", "suru verb - irregular" }, { "kyb", "Kyoto-ben" }, { "osb", "Osaka-ben" },
			{ "ksb", "Kansai-ben" }, { "ktb", "Kantou-ben" }, { "tsb", "Tosa-ben" }, { "thb", "Touhoku-ben" },
			{ "tsug", "Tsugaru-ben" }, { "kyu", "Kyuushuu-ben" }, { "rkb", "Ryuukyuu-ben" }, { "nab", "Nagano-ben" },
			{ "hob", "Hokkaido-ben" }, { "vt", "transitive verb" }, { "vulg", "vulgar expression or word" },
			{ "adj-kari", "`kari' adjective (archaic)" }, { "adj-ku", "`ku' adjective (archaic)" },
			{ "adj-shiku", "`shiku' adjective (archaic)" }, { "adj-nari", "archaic/formal form of na-adjective" },
			{ "n-pr", "proper noun" }, { "v-unspec", "verb unspecified" },
			{ "v4k", "Yodan verb with `ku' ending (archaic)" }, { "v4g", "Yodan verb with `gu' ending (archaic)" },
			{ "v4s", "Yodan verb with `su' ending (archaic)" }, { "v4t", "Yodan verb with `tsu' ending (archaic)" },
			{ "v4n", "Yodan verb with `nu' ending (archaic)" }, { "v4b", "Yodan verb with `bu' ending (archaic)" },
			{ "v4m", "Yodan verb with `mu' ending (archaic)" },
			{ "v2k-k", "Nidan verb (upper class) with `ku' ending (archaic)" },
			{ "v2g-k", "Nidan verb (upper class) with `gu' ending (archaic)" },
			{ "v2t-k", "Nidan verb (upper class) with `tsu' ending (archaic)" },
			{ "v2d-k", "Nidan verb (upper class) with `dzu' ending (archaic)" },
			{ "v2h-k", "Nidan verb (upper class) with `hu/fu' ending (archaic)" },
			{ "v2b-k", "Nidan verb (upper class) with `bu' ending (archaic)" },
			{ "v2m-k", "Nidan verb (upper class) with `mu' ending (archaic)" },
			{ "v2y-k", "Nidan verb (upper class) with `yu' ending (archaic)" },
			{ "v2r-k", "Nidan verb (upper class) with `ru' ending (archaic)" },
			{ "v2k-s", "Nidan verb (lower class) with `ku' ending (archaic)" },
			{ "v2g-s", "Nidan verb (lower class) with `gu' ending (archaic)" },
			{ "v2s-s", "Nidan verb (lower class) with `su' ending (archaic)" },
			{ "v2z-s", "Nidan verb (lower class) with `zu' ending (archaic)" },
			{ "v2t-s", "Nidan verb (lower class) with `tsu' ending (archaic)" },
			{ "v2d-s", "Nidan verb (lower class) with `dzu' ending (archaic)" },
			{ "v2n-s", "Nidan verb (lower class) with `nu' ending (archaic)" },
			{ "v2h-s", "Nidan verb (lower class) with `hu/fu' ending (archaic)" },
			{ "v2b-s", "Nidan verb (lower class) with `bu' ending (archaic)" },
			{ "v2m-s", "Nidan verb (lower class) with `mu' ending (archaic)" },
			{ "v2y-s", "Nidan verb (lower class) with `yu' ending (archaic)" },
			{ "v2r-s", "Nidan verb (lower class) with `ru' ending (archaic)" },
			{ "v2w-s", "Nidan verb (lower class) with `u' ending and `we' conjugation (archaic)" },
			{ "archit", "architecture term" }, { "astron", "astronomy, etc. term" }, { "baseb", "baseball term" },
			{ "biol", "biology term" }, { "bot", "botany term" }, { "bus", "business term" },
			{ "econ", "economics term" }, { "eng", "engineering term" }, { "fin", "finance term" },
			{ "geol", "geology, etc. term" }, { "law", "law, etc. term" }, { "med", "medicine, etc. term" },
			{ "music", "music term" }, { "Shinto", "Shinto term" }, { "sports", "sports term" },
			{ "sumo", "sumo term" }, { "zool", "zoology term" }, { "joc", "jocular, humorous term" }, };

	public static final HashSet<String> SET_POS;
	static {
		final int numPos = POS.length;
		SET_POS = new HashSet<String>(numPos);
		for (int i = 0; i < numPos; i++) {
			Sense.SET_POS.add(POS[i][0]);
		}
	}

	/**
	 * These elements, if present, indicate that the sense is restricted to the
	 * lexeme represented by the keb.
	 */
	public ArrayList<String> stagk;

	/**
	 * These elements, if present, indicate that the sense is restricted to the
	 * lexeme represented by the reb.
	 */
	public ArrayList<String> stagr;

	/**
	 * Part-of-speech information about the entry/sense. Should use appropriate
	 * entity codes. In general where there are multiple senses in an entry, the
	 * part-of-speech of an earlier sense will apply to later senses unless
	 * there is a new part-of-speech indicated.
	 */
	public HashSet<String> pos;

	/**
	 * This element is used to indicate a cross-reference to another entry with
	 * a similar or related meaning or sense. The content of this element is
	 * typically a keb or reb element in another entry. In some cases a keb will
	 * be followed by a reb and/or a sense number to provide a precise target
	 * for the cross-reference. Where this happens, a JIS "centre-dot" (0x2126)
	 * is placed between the components of the cross-reference.
	 */
	public ArrayList<String> xref;

	/**
	 * This element is used to indicate another entry which is an antonym of the
	 * current entry/sense. The content of this element must exactly match that
	 * of a keb or reb element in another entry.
	 */
	public ArrayList<String> ant;

	/**
	 * Information about the field of application of the entry/sense. When
	 * absent, general application is implied. Entity coding for specific fields
	 * of application.
	 */
	public ArrayList<String> field;

	/**
	 * This element is used for other relevant information about the
	 * entry/sense. As with part-of-speech, information will usually apply to
	 * several senses.
	 */
	public ArrayList<String> misc;

	/**
	 * The sense-information elements provided for additional information to be
	 * recorded about a sense. Typical usage would be to indicate such things as
	 * level of currency of a sense, the regional variations, etc.
	 */
	public ArrayList<String> s_inf;

	/**
	 * This element records the information about the source language(s) of a
	 * loan-word/gairaigo. If the source language is other than English, the
	 * language is indicated by the xml:lang attribute. The element value (if
	 * any) is the source word or phrase.
	 */
	public ArrayList<SenseLangSource> lsource;

	/**
	 * For words specifically associated with regional dialects in Japanese, the
	 * entity code for that dialect, e.g. ksb for Kansaiben.
	 */
	public ArrayList<String> dial;

	/**
	 * Within each sense will be one or more "glosses", i.e. target-language
	 * words or phrases which are equivalents to the Japanese word. This element
	 * would normally be present, however it may be omitted in entries which are
	 * purely for a cross-reference.
	 */
	public final ArrayList<SenseGloss> glosses;

	/**
	 * The example elements provide for pairs of short Japanese and
	 * target-language phrases or sentences which exemplify the usage of the
	 * Japanese head-word and the target-language gloss. Words in example fields
	 * would typically not be indexed by a dictionary application.
	 */
	public ArrayList<String> examples;

	public Sense() {
		this.stagk = null;
		this.stagr = null;
		this.pos = null;
		this.xref = null;
		this.ant = null;
		this.field = null;
		this.misc = null;
		this.s_inf = null;
		this.lsource = null;
		this.dial = null;
		this.glosses = new ArrayList<SenseGloss>();
		this.examples = null;
	}

	@Override
	public String toString() {
		// Useful when debugging
		final ArrayList<SenseGloss> lGlosses = this.glosses;
		final int nGlosses = lGlosses.size();
		final StringBuilder sb = new StringBuilder();
		final HashMap<String, Integer> glossPerLang = new HashMap<String, Integer>();
		for (int i = 0; i < nGlosses; i++) {
			SenseGloss sg = lGlosses.get(i);
			int val = 1;
			if (glossPerLang.containsKey(sg.lang)) {
				val = glossPerLang.get(sg.lang) + 1;
			}
			glossPerLang.put(sg.lang, val);
		}

		sb.append('[');
		for (String lang : glossPerLang.keySet()) {
			int val = glossPerLang.get(lang);
			sb.append(lang + ":" + val);
			sb.append(",");
		}
		sb.append(']');

		return sb.toString();
	}

}
